from audeer.core.config import config
from audeer.core.io import basename_wo_ext
from audeer.core.io import common_directory
from audeer.core.io import create_archive
from audeer.core.io import download_url
from audeer.core.io import extract_archive
from audeer.core.io import extract_archives
from audeer.core.io import file_extension
from audeer.core.io import list_dir_names
from audeer.core.io import list_file_names
from audeer.core.io import md5
from audeer.core.io import mkdir
from audeer.core.io import move
from audeer.core.io import move_file
from audeer.core.io import replace_file_extension
from audeer.core.io import rmdir
from audeer.core.io import script_dir
from audeer.core.io import touch
from audeer.core.path import path
from audeer.core.path import safe_path
from audeer.core.tqdm import format_display_message
from audeer.core.tqdm import progress_bar
from audeer.core.utils import deprecated
from audeer.core.utils import deprecated_default_value
from audeer.core.utils import deprecated_keyword_argument
from audeer.core.utils import flatten_list
from audeer.core.utils import freeze_requirements
from audeer.core.utils import git_repo_tags
from audeer.core.utils import git_repo_version
from audeer.core.utils import install_package
from audeer.core.utils import is_semantic_version
from audeer.core.utils import is_uid
from audeer.core.utils import run_tasks
from audeer.core.utils import run_worker_threads
from audeer.core.utils import sort_versions
from audeer.core.utils import to_list
from audeer.core.utils import uid
from audeer.core.utils import unique
from audeer.core.version import LooseVersion
from audeer.core.version import StrictVersion


# Discourage from audeer import *
__all__ = []


# Dynamically get the version of the installed module
try:
    import importlib.metadata

    __version__ = importlib.metadata.version(__name__)
except Exception:  # pragma: no cover
    importlib = None  # pragma: no cover
finally:
    del importlib



import audeer


# Project -----------------------------------------------------------------
project = "audeer"
author = "Hagen Wierstorf, Johannes Wagner"
version = audeer.git_repo_version()
title = "{} Documentation".format(project)


# General -----------------------------------------------------------------
master_doc = "index"
source_suffix = ".rst"
exclude_patterns = [
    "build",
    "tests",
    "Thumbs.db",
    ".DS_Store",
    "api-src",
]
pygments_style = None
extensions = [
    "sphinx.ext.napoleon",  # support for Google-style docstrings
    "sphinx_autodoc_typehints",
    "sphinx_copybutton",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx_apipages",
]
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
}
linkcheck_ignore = [
    "https://gitlab.audeering.com",
]
copybutton_prompt_text = r">>> |\.\.\. |$ "
copybutton_prompt_is_regexp = True

autodoc_default_options = {
    "undoc-members": False,
}
apipages_hidden_methods = [
    "__call__",
    "__eq__",
    "__lt__",
    "__le__",
    "__gt__",
    "__ge__",
    "__repr__",
    "__str__",
]


# HTML --------------------------------------------------------------------
html_theme = "sphinx_audeering_theme"
html_theme_options = {
    "display_version": True,
    "footer_links": False,
    "logo_only": False,
}
html_context = {
    "display_github": True,
}
html_title = title



import sybil
from sybil.parsers.rest import DocTestParser
from sybil.parsers.rest import PythonCodeBlockParser


pytest_collect_file = sybil.Sybil(
    parsers=[DocTestParser(), PythonCodeBlockParser()],
    pattern="*.rst",
).pytest()



from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Sequence
import concurrent.futures
import copy
import functools
import hashlib
import importlib
import importlib.metadata
import inspect
import multiprocessing
import operator
import queue
import subprocess
import sys
import threading
import uuid
import warnings

from audeer.core.tqdm import progress_bar as audeer_progress_bar
from audeer.core.version import LooseVersion


__doctest_skip__ = ["git_repo_tags", "git_repo_version"]


def deprecated(
    *,
    removal_version: str,
    alternative: str = None,
) -> Callable:
    r"""Mark code as deprecated.

    Provide a `decorator <https://www.python.org/dev/peps/pep-0318/>`_
    to mark functions/classes as deprecated.

    You have to specify the version,
    for which the deprecated code will be removed.
    If you change only small things
    like renaming a function or an argument,
    it will be fine to remove the code
    with the next minor release (`X.(Y+1).Z`).
    Otherwise, choose the next major release (`(X+1).Y.Z`).

    Args:
        removal_version: version the code will be removed
        alternative: alternative code to use

    Examples:
        >>> @audeer.deprecated(removal_version="2.0.0")
        ... def deprecated_function():
        ...     pass

    """

    def _deprecated(func):
        # functools.wraps preserves the name
        # and docstring of the decorated code:
        # https://docs.python.org/3/library/functools.html#functools.wraps
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            message = (
                f"{func.__name__} is deprecated and will be removed "
                f"with version {removal_version}."
            )
            if alternative is not None:
                message += f" Use {alternative} instead."
            warnings.warn(message, category=UserWarning, stacklevel=2)
            return func(*args, **kwargs)

        return new_func

    return _deprecated


def deprecated_default_value(
    *,
    argument: str,
    change_in_version: str,
    new_default_value: object,
) -> Callable:
    """Mark default value of keyword argument as deprecated.

    Provide a `decorator <https://www.python.org/dev/peps/pep-0318/>`_
    to mark the default value of a keyword argument as deprecated.
    You have to specify the version
    for which the default value will change
    and the new default value.

    Args:
        argument: keyword argument
        change_in_version: version the default value will change
        new_default_value: new default value

    Examples:
        >>> @audeer.deprecated_default_value(
        ...     argument="foo",
        ...     change_in_version="2.0.0",
        ...     new_default_value="bar",
        ... )
        ... def deprecated_function(foo="foo"):
        ...     pass
    """

    def _deprecated(func):
        # functools.wraps preserves the name
        # and docstring of the decorated code:
        # https://docs.python.org/3/library/functools.html#functools.wraps
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            if argument not in kwargs:
                signature = inspect.signature(func)
                default_value = signature.parameters[argument].default
                message = (
                    f"The default of '{argument}' will change from "
                    f"'{default_value}' to '{new_default_value}' "
                    f"with version {change_in_version}."
                )
                warnings.warn(message, category=UserWarning, stacklevel=2)
            return func(*args, **kwargs)

        return new_func

    return _deprecated


def deprecated_keyword_argument(
    *,
    deprecated_argument: str,
    removal_version: str,
    new_argument: str = None,
    mapping: Callable = None,
    remove_from_kwargs: bool = True,
) -> Callable:
    r"""Mark keyword argument as deprecated.

    Provide a `decorator <https://www.python.org/dev/peps/pep-0318/>`_
    to mark keyword arguments as deprecated.

    You have to specify the version,
    for which the deprecated argument will be removed.
    The content assigned to ``deprecated_argument``
    is passed on to the ``new_argument``.

    Args:
        deprecated_argument: keyword argument to be marked as deprecated
        removal_version: version the code will be removed
        new_argument: keyword argument that should be used instead
        mapping: if the keyword argument is not only renamed,
            but expects also different input values,
            you can map to the new ones with this callable
        remove_from_kwargs: if ``True``,
            ``deprecated_argument`` will be removed
            from ``kwargs`` inside the decorated object

    Examples:
        >>> @audeer.deprecated_keyword_argument(
        ...     deprecated_argument="foo",
        ...     new_argument="bar",
        ...     removal_version="2.0.0",
        ... )
        ... def function_with_new_argument(*, bar):
        ...     pass

    """

    def _deprecated(func):
        # functools.wraps preserves the name
        # and docstring of the decorated code:
        # https://docs.python.org/3/library/functools.html#functools.wraps
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            if deprecated_argument in kwargs:
                message = (
                    f"'{deprecated_argument}' argument is deprecated "
                    f"and will be removed with version {removal_version}."
                )
                if remove_from_kwargs:
                    argument_content = kwargs.pop(deprecated_argument)
                else:
                    argument_content = kwargs[deprecated_argument]
                if new_argument is not None:
                    message += f" Use '{new_argument}' instead."
                    if mapping is not None:
                        kwargs[new_argument] = mapping(argument_content)
                    else:
                        kwargs[new_argument] = argument_content
                warnings.warn(
                    message,
                    category=UserWarning,
                    stacklevel=2,
                )
            return func(*args, **kwargs)

        return new_func

    return _deprecated


def flatten_list(nested_list: list) -> list:
    """Flatten an arbitrarily nested list.

    Implemented without  recursion to avoid stack overflows.
    Returns a new list, the original list is unchanged.

    Args:
        nested_list: nested list

    Returns:
        flattened list

    Examples:
        >>> audeer.flatten_list([1, 2, 3, [4], [], [[[[[[[[[5]]]]]]]]]])
        [1, 2, 3, 4, 5]
        >>> audeer.flatten_list([[1, 2], 3])
        [1, 2, 3]

    """

    def _flat_generator(nested_list):
        while nested_list:
            sublist = nested_list.pop(0)
            if isinstance(sublist, list):
                nested_list = sublist + nested_list
            else:
                yield sublist

    nested_list = copy.deepcopy(nested_list)
    return list(_flat_generator(nested_list))


def freeze_requirements(outfile: str):
    r"""Log Python packages of activate virtual environment.

    Args:
        outfile: file to store the packages.
            Usually a name like :file:`requirements.txt.lock` should be picked.

    Raises:
        RuntimeError: if running ``pip freeze`` returns an error

    """
    cmd = f"pip freeze > {outfile}"
    with subprocess.Popen(
        args=cmd,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.PIPE,
        shell=True,
    ) as p:
        _, err = p.communicate()
        if bool(p.returncode):
            raise RuntimeError(f"Freezing Python packages failed: {err}")


def git_repo_tags(
    *,
    v: bool = None,
) -> list:
    r"""Get a list of available git tags.

    The tags are inferred by executing
    ``git tag`` in the current folder.
    If the command fails,
    an empty list is returned.

    Args:
        v: if ``True`` tags start always with ``v``,
            if ``False`` they never start with ``v``,
            if ``None`` the original tag names are returned

    Returns:
        list of tags

    Examples:
        .. skip: next

        >>> audeer.git_repo_tags()
        ['v1.0.0', 'v1.1.0', 'v2.0.0']

    """
    try:
        git = ["git", "tag"]
        tags = subprocess.check_output(git)
        tags = tags.decode().strip().split("\n")
    except Exception:  # pragma: no cover
        tags = []
    if v is None:
        return tags
    if v:
        tags = [f"v{t}" if not t.startswith("v") else t for t in tags]
    else:
        tags = [t.removeprefix("v") for t in tags]
    return tags


def git_repo_version(
    *,
    v: bool = True,
) -> str:
    r"""Get a version number from current git ref.

    The version is inferred executing
    ``git describe --tags --always``.
    If the command fails,
    ``'<unknown>'`` is returned.

    Args:
        v: if ``True`` version starts always with ``v``,
            otherwise it never starts with ``v``

    Returns:
        version number

    Examples:
        .. skip: next

        >>> audeer.git_repo_version()
        'v1.0.0'

    """
    try:
        git = ["git", "describe", "--tags", "--always"]
        version = subprocess.check_output(git)
        version = version.decode().strip()
    except Exception:  # pragma: no cover
        version = "<unknown>"
    if version.startswith("v") and not v:  # pragma: no cover (only local)
        version = version[1:]
    elif not version.startswith("v") and v:  # pragma: no cover (only github)
        version = f"v{version}"
    return version


def install_package(
    name: str,
    *,
    version: str = None,
    silent: bool = False,
):
    r"""Install a Python package with pip.

    An error is raised if a different version
    of the package is already installed.
    However,
    it is possible to use one of the following
    operators in front of the version string:
    ``'>='``, ``'>'``, ``'<='``, ``'<'``.
    In that case,
    an error is raised only
    if the condition is not satisfied.
    If version is set to ``None``
    and the package is not installed yet,
    the latest version will be installed.

    Args:
        name: package name
        version: version string (see description)
        silent: suppress messages to stdout

    Raises:
        subprocess.CalledProcessError: if the sub-process calling pip fails,
            e.g. because requested version of the package is not found
        RuntimeError: if a version of the package is already
            installed that does not satisfy the requested version

    """
    version = version.strip() if version is not None else None
    version_org = version
    op = operator.eq

    # check for operators, e.g.
    # >=1.0, >1.0, <=1.0, <1.0
    if version is not None:
        if version.startswith(">="):
            op = operator.ge
            version = version[2:].strip()
        elif version.startswith(">"):
            op = operator.gt
            version = version[1:].strip()
        elif version.startswith("<="):
            op = operator.le
            version = version[2:].strip()
        elif version.startswith("<"):
            op = operator.lt
            version = version[1:].strip()

    # raise error if package is already installed
    # and does not satisfy requested version
    try:
        current_version = importlib.metadata.version(name)
    except importlib.metadata.PackageNotFoundError:
        current_version = None

    if current_version is not None:
        if version is None:
            return  # any version is fine
        if op(LooseVersion(current_version), LooseVersion(version)):
            return  # installed version satisfies requested version
        raise RuntimeError(
            f"The installed version "
            f"'{current_version}' "
            f"of "
            f"{name} "
            "does not satisfy "
            f"'{version_org}'."
        )

    # install package
    version = version_org
    if version is not None:
        if op == operator.eq:
            # since we do not support ==1.0
            # we have to add it here
            name = f"{name}=={version}"
        else:
            name = f"{name}{version}"

    subprocess.check_call(
        [
            sys.executable,
            "-m",
            "pip",
            "install",
            name,
        ],
        stdout=subprocess.DEVNULL if silent else None,
    )
    # This function should be called if any modules
    # are created/installed while your program is running
    # to guarantee all finders will notice the new module’s existence.
    # see https://docs.python.org/3/library/importlib.html
    importlib.invalidate_caches()


def is_semantic_version(version: str) -> bool:
    r"""Check if given string represents a semantic version.

    To be a `semantic version`_
    your version has to comply to ``X.Y.Z`` or ``vX.Y.Z``,
    where X, Y, Z are all integers.
    Additional version information, like ``beta``
    has to be added using a ``-`` or ``+``,
    e.g. ``X.Y.Z-beta``.

    .. _semantic version: https://semver.org

    Args:
        version: version string

    Returns:
        ``True`` if version is a semantic version

    Examples:
        >>> audeer.is_semantic_version("v1")
        False
        >>> audeer.is_semantic_version("1.2.3-r3")
        True
        >>> audeer.is_semantic_version("v0.7.2-9-g1572b37")
        True

    """
    version_parts = version.split(".")

    if len(version_parts) < 3:
        return False

    def is_integer_convertable(x):
        try:
            int(x)
            return True
        except ValueError:
            return False

    x, y = version_parts[:2]
    # Ignore starting 'v'
    x = x.removeprefix("v")

    z = ".".join(version_parts[2:])
    # For Z, '-' and '+' are also allowed as separators,
    # but you are not allowed to have an additional '.' before
    z = z.split("-")[0]
    z = z.split("+")[0]
    if len(z.split(".")) > 1:
        return False

    for v in (x, y, z):
        if not is_integer_convertable(v):
            return False

    return True


def is_uid(uid: str) -> bool:
    r"""Check if string is a unique identifier.

    Unique identifiers can be generated with :func:`audeer.uid`.

    Args:
        uid: string

    Returns:
        ``True`` if string is a unique identifier

    Examples:
        >>> audeer.is_uid("626f68e6-d336-70b9-e753-ed9fad855840")
        True
        >>> audeer.is_uid("ad855840")
        True
        >>> audeer.is_uid("not a unique identifier")
        False

    """
    if uid is None:
        return False
    if not isinstance(uid, str):
        return False
    if len(uid) != 8 and len(uid) != 36:
        return False

    if len(uid) == 8:
        uid = f"00000000-0000-0000-0000-0000{uid}"

    for pos in [8, 13, 18, 23]:
        if not uid[pos] == "-":
            return False

    try:
        uuid.UUID(uid, version=1)
    except ValueError:
        return False

    return True


def run_tasks(
    task_func: Callable,
    params: Sequence[
        tuple[
            Sequence[object],
            dict[str, object],
        ]
    ],
    *,
    num_workers: int = 1,
    multiprocessing: bool = False,
    progress_bar: bool = False,
    task_description: str = None,
    maximum_refresh_time: float = None,
) -> list[object]:
    r"""Run parallel tasks using multprocessing.

    .. note:: Result values are returned in order of ``params``.

    Args:
        task_func: task function with one or more
            parameters, e.g. ``x, y, z``, and optionally returning a value
        params: sequence of tuples holding parameters for each task.
            Each tuple contains a sequence of positional arguments and a
            dictionary with keyword arguments, e.g.:
            ``[((x1, y1), {'z': z1}), ((x2, y2), {'z': z2}), ...]``
        num_workers: number of parallel jobs or 1 for sequential
            processing. If ``None`` will be set to the number of
            processors on the machine multiplied by 5 in case of
            multithreading and number of processors in case of
            multiprocessing
        multiprocessing: use multiprocessing instead of multithreading
        progress_bar: show a progress bar
        task_description: task description
            that will be displayed next to progress bar
        maximum_refresh_time: refresh the progress bar
            at least every ``maximum_refresh_time`` seconds,
            using another thread.
            If ``None``,
            no refreshing is enforced

    Returns:
        list of computed results

    Examples:
        >>> power = lambda x, n: x**n
        >>> params = [([2, n], {}) for n in range(10)]
        >>> audeer.run_tasks(power, params, num_workers=3)
        [1, 2, 4, 8, 16, 32, 64, 128, 256, 512]

    """
    num_tasks = max(1, len(params))
    results = [None] * num_tasks

    if num_workers == 1:  # sequential
        with audeer_progress_bar(
            params,
            total=len(params),
            desc=task_description,
            maximum_refresh_time=maximum_refresh_time,
            disable=not progress_bar,
        ) as pbar:
            for index, param in enumerate(pbar):
                results[index] = task_func(*param[0], **param[1])

    else:  # parallel
        if multiprocessing:
            executor = concurrent.futures.ProcessPoolExecutor
        else:
            executor = concurrent.futures.ThreadPoolExecutor
        with executor(max_workers=num_workers) as pool:
            with audeer_progress_bar(
                total=len(params),
                desc=task_description,
                maximum_refresh_time=maximum_refresh_time,
                disable=not progress_bar,
            ) as pbar:
                futures = []
                for param in params:
                    future = pool.submit(task_func, *param[0], **param[1])
                    future.add_done_callback(lambda p: pbar.update())
                    futures.append(future)
                for idx, future in enumerate(futures):
                    result = future.result()
                    results[idx] = result

    return results


@deprecated(removal_version="2.0.0", alternative="run_tasks")
def run_worker_threads(
    task_fun: Callable,
    params: Sequence[object] = None,
    *,
    num_workers: int = None,
    progress_bar: bool = False,
    task_description: str = None,
) -> Sequence[object]:  # pragma: no cover
    r"""Run parallel tasks using worker threads.

    .. note:: Result values are returned in order of ``params``.

    Args:
        task_fun: task function with one or more
            parameters, e.g. ``x, y, z``, and optionally returning a value
        params: list of parameters (use tuples in case of multiple parameters)
            for each task, e.g. ``[(x1, y1, z1), (x2, y2, z2), ...]``
        num_workers: number of worker threads (defaults to number of available
            CPUs multiplied by ``5``)
        progress_bar: show a progress bar
        task_description: task description
            that will be displayed next to progress bar

    Examples:
        >>> power = lambda x, n: x**n
        >>> params = [(2, n) for n in range(10)]
        >>> audeer.run_worker_threads(power, params, num_workers=3)
        [1, 2, 4, 8, 16, 32, 64, 128, 256, 512]

    """
    if params is None:
        params = []
    num_tasks = max(1, len(params))
    results = [None] * num_tasks
    if num_workers is None:
        num_workers = multiprocessing.cpu_count() * 5

    # Ensure num_workers is positive
    num_workers = max(1, num_workers)
    # Do not use more workers as needed
    num_workers = min(num_workers, num_tasks)

    # num_workers == 1 -> run sequentially
    if num_workers == 1:
        for index, param in enumerate(params):
            if type(param) in (list, tuple):
                results[index] = task_fun(*param)
            else:
                results[index] = task_fun(param)

    # num_workers > 1 -> run parallel
    else:
        # Create queue, possibly with a progress bar
        if progress_bar:

            class QueueWithProgbar(queue.Queue):
                def __init__(self, num_tasks, maxsize=0):
                    super().__init__(maxsize)
                    self.pbar = audeer_progress_bar(
                        total=num_tasks,
                        desc=task_description,
                    )

                def task_done(self):
                    super().task_done()
                    self.pbar.update(1)

            q = QueueWithProgbar(num_tasks)
        else:
            q = queue.Queue()

        # Fill queue
        for index, param in enumerate(params):
            q.put((index, param))

        # Define worker thread
        def _worker():
            while True:
                item = q.get()
                if item is None:
                    break
                index, param = item
                if type(param) in (list, tuple):
                    results[index] = task_fun(*param)
                else:
                    results[index] = task_fun(param)
                q.task_done()

        # Start workers
        threads = []
        for i in range(num_workers):
            t = threading.Thread(target=_worker)
            t.start()
            threads.append(t)

        # Block until all tasks are done
        q.join()

        # Stop workers
        for _ in range(num_workers):
            q.put(None)
        for t in threads:
            t.join()
    return results


def sort_versions(
    versions: list[str],
) -> list[str]:
    """Sort version numbers.

    If a version starts with ``v``,
    the ``v`` is ignored during sorting.

    Args:
        versions: sequence with semantic version numbers

    Returns:
        sorted list of versions with highest as last entry

    Raises:
        ValueError: if the version does not comply
            with :func:`is_semantic_version`

    Examples:
        >>> vers = [
        ...     "2.0.0",
        ...     "2.0.1",
        ...     "v1.0.0",
        ...     "v2.0.0-1-gdf29c4a",
        ... ]
        >>> audeer.sort_versions(vers)
        ['v1.0.0', '2.0.0', 'v2.0.0-1-gdf29c4a', '2.0.1']

    """
    for version in versions:
        if not is_semantic_version(version):
            raise ValueError(
                "All version numbers have to be semantic versions, "
                "following 'X.Y.Z', "
                "where X, Y, Z are integers. "
                f"But your version is: '{version}'."
            )

    def sort_key(value):
        value = value.removeprefix("v")
        return LooseVersion(value)

    return sorted(versions, key=sort_key)


def to_list(x: object):
    """Convert to list.

    If an iterable is passed,
    that is not a string it will be converted using :class:`list`.
    Otherwise, ``x`` is converted by ``[x]``.

    Args:
        x: input to be converted to a list

    Returns:
        input as a list

    Examples:
        >>> audeer.to_list("abc")
        ['abc']
        >>> audeer.to_list((1, 2, 3))
        [1, 2, 3]

    """
    if not isinstance(x, Iterable) or isinstance(x, str):
        return [x]
    else:
        return list(x)


def uid(
    *,
    from_string: str = None,
    short: bool = False,
) -> str:
    r"""Generate unique identifier.

    A unique identifier contains 36 characters
    with ``-`` at position 9, 14, 19, 24.
    If ``short`` is ``True``,
    only the last 8 digits are returned.

    Args:
        from_string: create a unique identifier
            by hashing the provided string.
            This will return the same identifier
            for identical strings.
            If ``None`` :func:`uuid.uuid1` is used.
        short: if ``True`` returns
            a short unique identifier
            (last 8 digits)

    Returns:
        unique identifier

    Examples:
        >>> audeer.uid(from_string="example_string")
        '626f68e6-d336-70b9-e753-ed9fad855840'
        >>> audeer.uid(from_string="example_string", short=True)
        'ad855840'

    """
    if from_string is None:
        uid = str(uuid.uuid1())
    else:
        uid = hashlib.md5()
        uid.update(from_string.encode("utf-8"))
        uid = uid.hexdigest()
        uid = f"{uid[0:8]}-{uid[8:12]}-{uid[12:16]}-{uid[16:20]}-{uid[20:]}"

    if short:
        uid = uid[-8:]

    return uid


def unique(sequence: Iterable) -> list:
    r"""Unique values in its original order.

    This is an alternative to ``list(set(x))``,
    which does not preserve the original order.

    Args:
        sequence: sequence of values

    Returns:
        unique values from ``x`` in order of appearance

    Examples:
        >>> list(set([2, 2, 1]))
        [1, 2]
        >>> audeer.unique([2, 2, 1])
        [2, 1]

    """
    # https://stackoverflow.com/a/480227
    seen = set()
    seen_add = seen.add
    return [x for x in sequence if not (x in seen or seen_add(x))]



import os

import pytest

import audeer


@pytest.mark.parametrize(
    "path",
    [
        ("~/.someconfigrc"),
        ("file.tar.gz"),
        ("/a/c.d/g"),
        ("/a/c.d/g.exe"),
        ("../.././README.md"),
        ("folder/file.txt"),
        (b"folder/file.txt"),
        (""),
    ],
)
def test_path(path):
    if path:
        expected_path = os.path.abspath(os.path.expanduser(path))
    else:
        expected_path = ""
    if isinstance(expected_path, bytes):
        expected_path = expected_path.decode("utf8")
    path = audeer.path(path)
    assert path == expected_path
    assert isinstance(path, str)
    path = audeer.safe_path(path)
    assert path == expected_path
    assert isinstance(path, str)


@pytest.mark.parametrize(
    "path, paths",
    [
        ("/a/b", ["file.tar.gz"]),
        ("/a", ["b", "file.tar.gz"]),
        ("/a/c.d/g", ["../f"]),
        ("", ""),
    ],
)
def test_path_join(path, paths):
    expected_path = os.path.join(path, *paths)
    if expected_path:
        expected_path = os.path.abspath(os.path.expanduser(expected_path))
    else:
        expected_path = ""
    path = audeer.path(path, *paths)
    assert path == expected_path
    assert isinstance(path, str)


def test_path_symlinks(tmpdir):
    filename = "file.txt"
    linkname = "link.txt"
    dir_tmp = tmpdir.mkdir("folder")
    f = dir_tmp.join(filename)
    f.write("")
    folder = audeer.mkdir(str(dir_tmp))
    file = os.path.join(folder, filename)
    link = os.path.join(folder, linkname)
    os.symlink(file, link)
    for follow_symlink in [True, False]:
        if follow_symlink:
            expected_path = os.path.realpath(os.path.expanduser(link))
        else:
            expected_path = os.path.abspath(os.path.expanduser(link))
        path = audeer.path(link, follow_symlink=follow_symlink)
    _, path = os.path.splitdrive(path)
    _, expected_path = os.path.splitdrive(expected_path)
    assert path == expected_path
    assert isinstance(path, str)



from __future__ import annotations

from collections.abc import Sequence
import os


def path(
    path: str | bytes,
    *paths: Sequence[str | bytes],
    follow_symlink: bool = False,
) -> str:
    """Expand and normalize to absolute path.

    It uses :func:`os.path.realpath`
    and :func:`os.path.expanduser`
    to ensure an absolute path
    without ``..`` or ``~``,
    and independent of the path separator
    of the operating system.
    If ``follow_symlink`` is ``False``,
    the faster :func:`os.path.abspath` is used
    instead of :func:`os.path.realpath`.

    Args:
        path: path to file, directory
        *paths: additional arguments
            to be joined with ``path``
            by :func:`os.path.join`
        follow_symlink: if ``True``
            symlinks are followed
            and the path of the original file
            is returned

    Returns:
        (joined and) expanded path

    Examples:
        .. skip: start if(platform.system() == "Windows")

        >>> home = audeer.path("~")
        >>> folder = audeer.path("~/path/.././path")
        >>> folder[len(home) + 1 :]
        'path'
        >>> file = audeer.path("~/path/.././path", "./file.txt")
        >>> file[len(home) + 1 :]
        'path/file.txt'
        >>> file = audeer.touch("file.txt")
        >>> link = audeer.path("link.txt")
        >>> os.symlink(file, link)
        >>> os.path.basename(audeer.path(link))
        'link.txt'
        >>> os.path.basename(audeer.path(link, follow_symlink=True))
        'file.txt'

        .. skip: end

    """
    if paths:
        path = os.path.join(path, *paths)
    if path:
        path = os.path.expanduser(path)
        if follow_symlink:
            path = os.path.realpath(path)
        else:
            path = os.path.abspath(path)
        # Convert bytes to str, see https://stackoverflow.com/a/606199
        if isinstance(path, bytes):
            path = path.decode("utf-8").strip("\x00")
    return path


# Ensure function is not hidden
# by `path` argument in `safe_path()`
_path = path


def safe_path(
    path: str | bytes,
    *paths: Sequence[str | bytes],
    follow_symlink: bool = False,
) -> str:
    """Expand and normalize to absolute path.

    It uses :func:`os.path.realpath`
    and :func:`os.path.expanduser`
    to ensure an absolute path
    without ``..`` or ``~``,
    and independent of the path separator
    of the operating system.
    If ``follow_symlink`` is ``False``,
    the faster :func:`os.path.abspath` is used
    instead of :func:`os.path.realpath`.

    Warning:
        :func:`audeer.safe_path` is deprecated,
        please use :func:`audeer.path` instead.

    Args:
        path: path to file, directory
        *paths: additional arguments
            to be joined with ``path``
            by :func:`os.path.join`
        follow_symlink: if ``True``
            symlinks are followed
            and the path of the original file
            is returned

    Returns:
        (joined and) expanded path

    Examples:
        .. skip: start if(platform.system() == "Windows")

        >>> home = audeer.safe_path("~")
        >>> folder = audeer.safe_path("~/path/.././path")
        >>> folder[len(home) + 1 :]
        'path'
        >>> file = audeer.safe_path("~/path/.././path", "./file.txt")
        >>> file[len(home) + 1 :]
        'path/file.txt'
        >>> file = audeer.touch("file.txt")
        >>> link = audeer.path("link.txt")
        >>> os.symlink(file, link)
        >>> os.path.basename(audeer.path(link))
        'link.txt'
        >>> os.path.basename(audeer.path(link, follow_symlink=True))
        'file.txt'

        .. skip: end

    """
    return _path(path, *paths, follow_symlink=follow_symlink)



import os
import subprocess
import warnings

import pytest

import audeer


def test_deprecated():
    @audeer.deprecated(removal_version="1.0.0", alternative="audeer.mkdir")
    def deprecated_function():
        pass

    expected_message = (
        "deprecated_function is deprecated "
        "and will be removed with version 1.0.0."
        " Use audeer.mkdir instead."
    )
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Raise warning
        deprecated_function()
        assert issubclass(w[-1].category, UserWarning)
        assert expected_message == str(w[-1].message)

    @audeer.deprecated(removal_version="2.0.0")
    def deprecated_function():
        pass

    expected_message = (
        "deprecated_function is deprecated " "and will be removed with version 2.0.0."
    )
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Raise warning
        deprecated_function()
        assert issubclass(w[-1].category, UserWarning)
        assert expected_message == str(w[-1].message)

    @audeer.deprecated(removal_version="2.0.0")
    class DeprecatedClass:
        pass

    expected_message = (
        "DeprecatedClass is deprecated " "and will be removed with version 2.0.0."
    )
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Raise warning
        DeprecatedClass()
        assert issubclass(w[-1].category, UserWarning)
        assert expected_message == str(w[-1].message)


def test_deprecated_default_value():
    @audeer.deprecated_default_value(
        argument="foo",
        new_default_value="bar",
        change_in_version="1.0.0",
    )
    def function_with_deprecated_default_value(*, foo="foo"):
        return foo

    expected_message = (
        "The default of 'foo' will change from " "'foo' to 'bar' " "with version 1.0.0."
    )
    default_value = "foo"
    with warnings.catch_warnings():
        # no warning if we set value
        warnings.simplefilter("error")
        function_with_deprecated_default_value(foo="foo")
        function_with_deprecated_default_value(foo="bar")
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Raise warning
        assert function_with_deprecated_default_value() == default_value
        assert issubclass(w[-1].category, UserWarning)
        assert expected_message == str(w[-1].message)


def test_deprecated_keyword_argument():
    @audeer.deprecated_keyword_argument(
        deprecated_argument="foo",
        new_argument="bar",
        removal_version="1.0.0",
        mapping=lambda x: 2 * x,
    )
    def function_with_deprecated_keyword_argument(*, bar):
        return bar

    expected_message = (
        "'foo' argument is deprecated "
        "and will be removed with version 1.0.0."
        " Use 'bar' instead."
    )
    value = 1
    assert function_with_deprecated_keyword_argument(bar=value) == value
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Raise warning
        r = function_with_deprecated_keyword_argument(foo=value)
        assert issubclass(w[-1].category, UserWarning)
        assert expected_message == str(w[-1].message)
        assert r == 2 * value

    @audeer.deprecated_keyword_argument(
        deprecated_argument="foo",
        removal_version="1.0.0",
    )
    def function_with_deprecated_keyword_argument(**kwargs):
        return 1

    expected_message = (
        "'foo' argument is deprecated " "and will be removed with version 1.0.0."
    )
    assert function_with_deprecated_keyword_argument() == 1
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Raise warning
        r = function_with_deprecated_keyword_argument(foo=2)
        assert issubclass(w[-1].category, UserWarning)
        assert expected_message == str(w[-1].message)
        assert r == 1

    @audeer.deprecated_keyword_argument(
        deprecated_argument="foo",
        removal_version="1.0.0",
        remove_from_kwargs=False,
    )
    def function_with_deprecated_keyword_argument(**kwargs):
        if "foo" in kwargs:
            return kwargs["foo"]
        else:
            return 1

    expected_message = (
        "'foo' argument is deprecated " "and will be removed with version 1.0.0."
    )
    assert function_with_deprecated_keyword_argument() == 1
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Raise warning
        r = function_with_deprecated_keyword_argument(foo=2)
        assert issubclass(w[-1].category, UserWarning)
        assert expected_message == str(w[-1].message)
        assert r == 2

    @audeer.deprecated_keyword_argument(
        deprecated_argument="foo",
        new_argument="bar",
        removal_version="1.0.0",
        remove_from_kwargs=False,
    )
    def function_with_deprecated_keyword_argument(**kwargs):
        if "foo" in kwargs:
            return kwargs["foo"]
        else:
            return 1

    expected_message = (
        "'foo' argument is deprecated "
        "and will be removed with version 1.0.0."
        " Use 'bar' instead."
    )
    assert function_with_deprecated_keyword_argument() == 1
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Raise warning
        r = function_with_deprecated_keyword_argument(foo=2)
        assert issubclass(w[-1].category, UserWarning)
        assert expected_message == str(w[-1].message)
        assert r == 2

    @audeer.deprecated_keyword_argument(
        deprecated_argument="foo",
        new_argument="bar",
        removal_version="1.0.0",
    )
    class ClassWithDeprecatedKeywordArgument(object):
        def __init__(self, *, bar):
            self.bar = bar

    expected_message = (
        "'foo' argument is deprecated "
        "and will be removed with version 1.0.0."
        " Use 'bar' instead."
    )
    value = 1
    assert ClassWithDeprecatedKeywordArgument(bar=value).bar == value
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Raise warning
        r = ClassWithDeprecatedKeywordArgument(foo=value)
        assert issubclass(w[-1].category, UserWarning)
        assert expected_message == str(w[-1].message)
        assert r.bar == value


@pytest.mark.parametrize(
    "nested_list,expected_list",
    [
        ([1, 2, 3, [4], [], [[[[[[[[[5]]]]]]]]]], [1, 2, 3, 4, 5]),
        ([[1, 2], 3], [1, 2, 3]),
        ([1, 2, 3], [1, 2, 3]),
    ],
)
def test_flatten_list(nested_list, expected_list):
    flattened_list = audeer.flatten_list(nested_list)
    assert flattened_list == expected_list


def test_freeze_requirements(tmpdir):
    path = str(tmpdir.mkdir("tmp"))
    outfile = os.path.join(path, "requirements.txt.lock")
    audeer.freeze_requirements(outfile)
    with open(outfile) as f:
        requirements = f.readlines()
    # Remove whitespace and \n
    requirements = [r.strip() for r in requirements]
    assert any(["pytest" in r for r in requirements])
    assert any(["audeer" in r for r in requirements])
    with pytest.raises(RuntimeError, match=r"Freezing Python packages failed"):
        outfile = os.path.join(path, "tmp2/requirements.txt.lock")
        audeer.freeze_requirements(outfile)


def test_git_repo_tags():
    git = ["git", "tag"]
    expected_tags = subprocess.check_output(git)
    expected_tags = expected_tags.decode().strip().split("\n")
    tags = audeer.git_repo_tags()
    assert tags == expected_tags
    tags = audeer.git_repo_tags(v=True)
    expected_tags = [f"v{t}" if not t.startswith("v") else t for t in expected_tags]
    assert tags == expected_tags
    tags = audeer.git_repo_tags(v=False)
    expected_tags = [t[1:] if t.startswith("v") else t for t in expected_tags]
    assert tags == expected_tags


def test_git_repo_version():
    git = ["git", "describe", "--tags", "--always"]
    expected_version = subprocess.check_output(git)
    expected_version = expected_version.decode().strip()
    version = audeer.git_repo_version(v=True)
    if not expected_version.startswith("v"):
        expected_version = f"v{expected_version}"
    assert version == expected_version
    version = audeer.git_repo_version(v=False)
    assert version == expected_version[1:]


@pytest.mark.parametrize(
    "version, is_semantic",
    [
        (
            "1.0.0",
            True,
        ),
        (
            "v1.0.0",
            True,
        ),
        (
            "4.0.0-20200206.095534-3",
            True,
        ),
        (
            "v1.0.1-1-gdf29c4a",
            True,
        ),
        (
            "1",
            False,
        ),
        (
            "v1",
            False,
        ),
        (
            "v1.3-r3",
            False,
        ),
        (
            "v1.3.3-r3",
            True,
        ),
        (
            "a.b.c",
            False,
        ),
        (
            "va.b.c",
            False,
        ),
        (
            "1.2.a",
            False,
        ),
        (
            "v1.3.3.3-r3",
            False,
        ),
        (
            "1.0.0+20130313144700",
            True,
        ),
        (
            "1.0.0-alpha+001",
            True,
        ),
    ],
)
def test_is_semantic_version(version, is_semantic):
    assert audeer.is_semantic_version(version) == is_semantic


@pytest.mark.parametrize(
    "uid, expected",
    [
        (audeer.uid(), True),
        (audeer.uid(short=True), True),
        (audeer.uid(from_string="from string"), True),
        (audeer.uid(from_string="from string", short=True), True),
        (audeer.uid(from_string="from string", short=True).upper(), True),
        (None, False),
        (1234, False),
        ("", False),
        ("some random string", False),
        (audeer.uid()[:-1], False),
        (audeer.uid(short=True)[:-1], False),
        ("00000000-0000-0000-0000-000000000000", True),
        ("000000000-0000-0000-0000-00000000000", False),
        ("?0000000-0000-0000-0000-000000000000", False),
    ],
)
def test_is_uid(uid, expected):
    assert audeer.is_uid(uid) == expected


def power(a: int = 0, *, b: int = 1):
    return a**b


@pytest.mark.parametrize(
    "multiprocessing",
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    "num_workers",
    [
        1,
        3,
        None,
    ],
)
@pytest.mark.parametrize(
    "task_fun, params",
    [
        (power, [([], {})]),
        (power, [([1], {})]),
        (power, [([1], {"b": 2})]),
        (power, [([], {"a": 1, "b": 2})]),
        (power, [([x], {"b": x}) for x in range(5)]),
    ],
)
def test_run_tasks(multiprocessing, num_workers, task_fun, params):
    expected = [task_fun(*param[0], **param[1]) for param in params]
    results = audeer.run_tasks(
        task_fun,
        params,
        num_workers=num_workers,
        multiprocessing=multiprocessing,
    )
    assert expected == results


@pytest.mark.parametrize(
    "func,params,expected_output",
    [
        (
            lambda x, n: x**n,
            [(2, n) for n in range(7)],
            [1, 2, 4, 8, 16, 32, 64],
        ),
        (3, None, [None]),
        (lambda x: x, ["hello"], ["hello"]),
        (lambda x: x, ["hello", 1, print], ["hello", 1, print]),
    ],
)
def test_run_worker_threads(func, params, expected_output):
    num_workers = [None, 1, 2, 4, 5, 100]
    for n in num_workers:
        with pytest.warns(UserWarning):
            output = audeer.run_worker_threads(func, params, num_workers=n)
        assert output == expected_output


@pytest.mark.parametrize(
    "versions, expected_versions",
    [
        (
            ["1.0.0", "3.0.0", "1.10.1", "1.2.0", "1.0.1"],
            ["1.0.0", "1.0.1", "1.2.0", "1.10.1", "3.0.0"],
        ),
        (
            [
                "1.0.0-SNAPSHOT",
                "4.0.0-20200206.095424-2",
                "1.0.0",
                "2.0.0-20200131.102442-1",
                "3.0.0",
                "3.1.0",
                "4.0.0-20200206.095316-1",
                "3.2.0",
                "2.0.0-20200131.102728-2",
                "3.3.0",
                "3.4.0",
                "4.0.0-20200206.095534-3",
                "4.0.0",
            ],
            [
                "1.0.0",
                "1.0.0-SNAPSHOT",
                "2.0.0-20200131.102442-1",
                "2.0.0-20200131.102728-2",
                "3.0.0",
                "3.1.0",
                "3.2.0",
                "3.3.0",
                "3.4.0",
                "4.0.0",
                "4.0.0-20200206.095316-1",
                "4.0.0-20200206.095424-2",
                "4.0.0-20200206.095534-3",
            ],
        ),
        (
            ["v1.0.0", "v1.0.1", "v1.0.1-1-gdf29c4a"],
            ["v1.0.0", "v1.0.1", "v1.0.1-1-gdf29c4a"],
        ),
        # From https://github.com/postmarketOS/pmbootstrap/issues/342
        (
            ["22.7.3-r1.3", "22.7.3-r1"],
            ["22.7.3-r1", "22.7.3-r1.3"],
        ),
        (
            ["1.0.0", "1.1.1+1", "1.2.1", "1.2.0"],
            ["1.0.0", "1.1.1+1", "1.2.0", "1.2.1"],
        ),
    ],
)
def test_sort_versions(versions, expected_versions):
    sorted_versions = audeer.sort_versions(versions)
    assert sorted_versions == expected_versions


@pytest.mark.parametrize(
    "versions, error_message",
    [
        (
            ["1"],
            (
                "All version numbers have to be semantic versions, "
                "following 'X.Y.Z', "
                "where X, Y, Z are integers. "
                "But your version is: '1'."
            ),
        ),
    ],
)
def test_sort_versions_errors(versions, error_message):
    with pytest.raises(ValueError, match=error_message):
        audeer.sort_versions(versions)


@pytest.mark.parametrize(
    "input,expected_output",
    [
        (1, [1]),
        ("abc", ["abc"]),
        (["abc", 1], ["abc", 1]),
        ((1, 2, 3), [1, 2, 3]),
        (len, [len]),
        ({1: "a", 2: "b"}, [1, 2]),
        ([], []),
        ("", [""]),
        (None, [None]),
    ],
)
def test_to_list(input, expected_output):
    assert audeer.to_list(input) == expected_output


@pytest.mark.parametrize(
    "short",
    [
        False,
        True,
    ],
)
@pytest.mark.parametrize(
    "from_string",
    [
        None,
        "example_string",
    ],
)
def test_uid(from_string, short):
    uid = audeer.uid(from_string=from_string, short=short)
    if short:
        assert len(uid) == 8
    else:
        assert len(uid) == 36
        for pos in [8, 13, 18, 23]:
            assert uid[pos] == "-"
        uid2 = audeer.uid(from_string=from_string, short=short)
        if from_string is not None:
            assert uid == uid2
        else:
            assert uid != uid2


@pytest.mark.parametrize(
    "sequence, expected",
    [
        ([], []),
        ([1], [1]),
        ([2, 1], [2, 1]),
        ([2, 1, 2], [2, 1]),
        (["a", 1, "a", 1], ["a", 1]),
        ((1, 1), [1]),
        ("dddnnfhg", ["d", "n", "f", "h", "g"]),
        ([None, None], [None]),
    ],
)
def test_unique(sequence, expected):
    r"""Test audeer.unique().

    Should return unique values in original order.

    Args:
        sequence: sequence if input values
        expected: expected unique list

    """
    assert audeer.unique(sequence) == expected



from collections.abc import Sequence
import threading
import time

from tqdm import tqdm

from audeer.core.config import config


def format_display_message(text: str, pbar: bool = False) -> str:
    """Ensure a fixed length of text printed to screen.

    The length of the text message is the same as the overall
    progress bar length as given by :attr:`audeer.config.TQDM_COLUMNS`.

    Args:
        text: text to be displayed
        pbar: if a progress bar is displayed as well.
            This will shorten the text to the given progress bar
            description length :attr:`audeer.config.TQDM_DESCLEN`

    Returns:
        formatted text message

    Examples:
        >>> audeer.config.TQDM_COLUMNS = 20
        >>> audeer.format_display_message("Long text that will be shorten to fit")
        'Long te...n to fit'

    """
    if not text:
        return text
    if pbar:
        n = config.TQDM_DESCLEN - 2
    else:
        n = config.TQDM_COLUMNS - 2
    if len(text) == n:
        return text
    elif len(text) < n:
        return text.ljust(n)
    else:
        m = (n - 3) // 2
        return f"{text[:m]}...{text[len(text) - (n - m - 3):]}"


def progress_bar(
    iterable: Sequence = None,
    *,
    total: int = None,
    desc: str = None,
    disable: bool = False,
    maximum_refresh_time: float = None,
) -> tqdm:
    r"""Progress bar with optional text on the right.

    If you want to show a constant description text
    during presenting the progress bar,
    you can use it similar to:

    .. code-block:: python

        for file in progress_bar(files, desc="Copying"):
            copy(file)

    When the text should be updated as well,
    you have to explicitly do that in each step:

    .. code-block:: python

        with progress_bar(files) as pbar:
            for file in pbar:
                desc = format_display_message(
                    f"Copying {file}",
                    pbar=True,
                )
                pbar.set_description_str(desc)
                pbar.refresh()
                copy(file)
                pbar.update()

    Args:
        iterable: sequence to iterate through
        total: total number of iterations
        desc: text shown on the right of the progress bar
        disable: don't show the display bar
        maximum_refresh_time: refresh the progress bar
            at least every ``maximum_refresh_time`` seconds,
            using another thread.
            If ``None``,
            no refreshing is enforced

    Returns:
        progress bar object

    """
    if desc is None:
        desc = ""
    return tqdm_wrapper(
        iterable=iterable,
        maximum_refresh_time=maximum_refresh_time,
        ncols=config.TQDM_COLUMNS,
        bar_format=config.TQDM_FORMAT,
        total=total,
        disable=disable,
        desc=format_display_message(desc, pbar=True),
        leave=config.TQDM_LEAVE,
        smoothing=0,
    )


def tqdm_wrapper(
    iterable: Sequence,
    maximum_refresh_time: float,
    *args,
    **kwargs,
) -> tqdm:
    r"""Tqdm progress bar wrapper to enforce update once a second.

    When using tqdm with large time durations
    between single steps of the iteration,
    it will not automatically update the elapsed time,
    but needs to be forced,
    see https://github.com/tqdm/tqdm/issues/861#issuecomment-2197893883.

    Args:
        iterable: sequence to iterate through
        maximum_refresh_time: refresh the progress bar
            at least every ``maximum_refresh_time`` seconds,
            using another thread.
            If ``None``,
            no refreshing is enforced
        args: arguments passed on to ``tqdm``
        kwargs: keyword arguments passed on to ``tqdm``

    Returns:
        progress bar object

    """
    pbar = tqdm(iterable, *args, **kwargs)

    def refresh():
        while not pbar.disable:
            time.sleep(maximum_refresh_time)
            pbar.refresh()

    if maximum_refresh_time is not None:
        thread = threading.Thread(target=refresh, daemon=True)
        thread.start()

    return pbar



from doctest import ELLIPSIS
import os
import platform

import pytest
import sybil
from sybil.parsers.rest import DocTestParser
from sybil.parsers.rest import SkipParser

import audeer


def imports(namespace):
    """Provide Python modules to namespace."""
    namespace["platform"] = platform
    namespace["audeer"] = audeer


@pytest.fixture(scope="function")
def run_in_tmpdir(tmpdir_factory):
    """Move to a persistent tmpdir for execution of a whole file."""
    tmpdir = tmpdir_factory.mktemp("tmp")
    current_dir = os.getcwd()
    os.chdir(tmpdir)

    yield

    os.chdir(current_dir)


# Collect doctests
pytest_collect_file = sybil.Sybil(
    parsers=[DocTestParser(optionflags=ELLIPSIS), SkipParser()],
    patterns=["*.py"],
    fixtures=["run_in_tmpdir"],
    setup=imports,
).pytest()



import os
import platform
import stat
import sys
import time

import pytest

import audeer


@pytest.fixture(scope="function", autouse=False)
def tree(tmpdir, request):
    r"""Create file tree."""
    files = request.param
    paths = []

    for path in files:
        if os.name == "nt":
            path = path.replace("/", os.path.sep)
        if path.endswith(os.path.sep):
            path = audeer.path(tmpdir, path)
            path = audeer.mkdir(path)
            path = path + os.path.sep
            paths.append(path)
        else:
            path = audeer.path(tmpdir, path)
            audeer.mkdir(os.path.dirname(path))
            path = audeer.touch(path)
            paths.append(path)

    yield paths


@pytest.mark.parametrize(
    "tree, root, files, archive_create, archive_extract, destination, " "expected",
    [
        (  # empty
            [],
            ".",
            [],
            "archive.zip",
            "archive.zip",
            ".",
            [],
        ),
        (
            [],
            ".",
            None,
            "archive.zip",
            "archive.zip",
            ".",
            [],
        ),
        (  # single file
            ["file.txt"],
            ".",
            "file.txt",
            "archive.zip",
            "archive.zip",
            ".",
            ["file.txt"],
        ),
        (  # archive folder does not exist
            ["file.txt"],
            ".",
            "file.txt",
            "sub/archive.zip",
            "sub/archive.zip",
            ".",
            ["file.txt"],
        ),
        (  # file in sub folder
            ["file.txt", "sub/a/b/file.txt"],
            ".",
            ["sub/a/b/file.txt", "file.txt"],
            "archive.zip",
            "archive.zip",
            ".",
            ["sub/a/b/file.txt", "file.txt"],
        ),
        (
            ["file.txt", "sub/a/b/file.txt"],
            ".",
            ["sub/a/b/file.txt"],
            "archive.zip",
            "archive.zip",
            ".",
            ["sub/a/b/file.txt"],
        ),
        (
            ["file.txt", "sub/a/b/file.txt"],
            ".",
            ["file.txt"],
            "archive.zip",
            "archive.zip",
            ".",
            ["file.txt"],
        ),
        (  # include all files
            ["file.txt", "sub/a/b/file.txt", ".hidden"],
            ".",
            None,
            "archive.zip",
            "archive.zip",
            ".",
            [".hidden", "file.txt", "sub/a/b/file.txt"],
        ),
        (  # exclude empty folder
            ["file.txt", "sub/a/b/file.txt", ".hidden", "empty/"],
            ".",
            None,
            "archive.zip",
            "archive.zip",
            ".",
            [".hidden", "file.txt", "sub/a/b/file.txt"],
        ),
        (  # tar
            ["file.txt", "sub/a/b/file.txt"],
            ".",
            ["sub/a/b/file.txt", "file.txt"],
            "archive.tar",
            "archive.tar",
            ".",
            ["sub/a/b/file.txt", "file.txt"],
        ),
        (  # tar.gz
            ["file.txt", "sub/a/b/file.txt"],
            ".",
            ["sub/a/b/file.txt", "file.txt"],
            "archive.tar.gz",
            "archive.tar.gz",
            ".",
            ["sub/a/b/file.txt", "file.txt"],
        ),
        (  # tar.bz2
            ["file.txt", "sub/a/b/file.txt"],
            ".",
            ["sub/a/b/file.txt", "file.txt"],
            "archive.tar.bz2",
            "archive.tar.bz2",
            ".",
            ["sub/a/b/file.txt", "file.txt"],
        ),
        (  # tar.xz
            ["file.txt", "sub/a/b/file.txt"],
            ".",
            ["sub/a/b/file.txt", "file.txt"],
            "archive.tar.xz",
            "archive.tar.xz",
            ".",
            ["sub/a/b/file.txt", "file.txt"],
        ),
        (  # root is sub folder
            ["sub/file.txt"],
            "./sub",
            ["file.txt"],
            "archive.zip",
            "archive.zip",
            ".",
            ["file.txt"],
        ),
        (
            ["sub/file.txt"],
            "./sub",
            None,
            "archive.zip",
            "archive.zip",
            ".",
            ["file.txt"],
        ),
        (  # destitation is sub folder
            ["file.txt"],
            ".",
            ["file.txt"],
            "archive.zip",
            "archive.zip",
            "./sub",
            ["file.txt"],
        ),
        (  # relative path with ../
            ["sub/file.txt"],
            "./sub",
            ["../sub/file.txt"],
            "archive.zip",
            "archive.zip",
            ".",
            ["file.txt"],
        ),
        pytest.param(  # root does not exit
            [],
            "./sub",
            None,
            "archive.zip",
            "archive.zip",
            ".",
            None,
            marks=pytest.mark.xfail(raises=FileNotFoundError),
        ),
        pytest.param(  # root is not a directory
            ["file.txt"],
            "file.txt",
            None,
            "archive.zip",
            "archive.zip",
            ".",
            None,
            marks=pytest.mark.xfail(raises=NotADirectoryError),
        ),
        pytest.param(  # destination is not a directory
            ["file.txt"],
            ".",
            [],
            "archive.zip",
            "archive.zip",
            "file.txt",
            [],
            marks=pytest.mark.xfail(raises=NotADirectoryError),
        ),
        pytest.param(  # archive to be extracted does not exit
            [],
            ".",
            [],
            "archive.zip",
            "bad.zip",
            ".",
            None,
            marks=pytest.mark.xfail(raises=FileNotFoundError),
        ),
        pytest.param(  # archive to be extracted is a directory
            ["sub/"],
            ".",
            [],
            "archive.zip",
            "sub",
            ".",
            None,
            marks=pytest.mark.xfail(raises=IsADirectoryError),
        ),
        pytest.param(  # file does not exit
            [],
            ".",
            ["file.txt"],
            "archive.zip",
            "archive.zip",
            ".",
            None,
            marks=pytest.mark.xfail(raises=FileNotFoundError),
        ),
        pytest.param(  # file not below root
            ["file.txt", "sub/"],
            "./sub",
            ["../file.txt"],
            "archive.zip",
            "archive.zip",
            ".",
            None,
            marks=pytest.mark.xfail(raises=RuntimeError),
        ),
        pytest.param(  # archive type not supported
            [],
            ".",
            [],
            "archive.bad",
            "archive.zip",
            ".",
            None,
            marks=pytest.mark.xfail(raises=RuntimeError),
        ),
        pytest.param(
            ["archive.bad"],
            ".",
            [],
            "archive.zip",
            "archive.bad",
            ".",
            None,
            marks=pytest.mark.xfail(raises=RuntimeError),
        ),
        pytest.param(  # invalid .rar format
            ["file.txt", "sub/a/b/file.txt"],
            ".",
            ["sub/a/b/file.txt", "file.txt"],
            "archive.rar",
            "archive.rar",
            ".",
            None,
            marks=pytest.mark.xfail(raises=RuntimeError),
        ),
        pytest.param(  # invalid .7z format
            ["file.txt", "sub/a/b/file.txt"],
            ".",
            ["sub/a/b/file.txt", "file.txt"],
            "archive.7z",
            "archive.7z",
            ".",
            None,
            marks=pytest.mark.xfail(raises=RuntimeError),
        ),
        pytest.param(  # broken archive
            ["archive.zip"],
            ".",
            [],
            "archive.tar.gz",
            "archive.zip",
            ".",
            None,
            marks=pytest.mark.xfail(raises=RuntimeError),
        ),
    ],
    indirect=["tree"],
)
def test_archives(
    tmpdir, tree, root, files, archive_create, archive_extract, destination, expected
):
    root = audeer.path(tmpdir, root)
    destination = audeer.path(tmpdir, destination)
    archive_create = audeer.path(tmpdir, archive_create)
    archive_extract = audeer.path(tmpdir, archive_extract)

    if os.name == "nt":
        if expected is not None:
            expected = [file.replace("/", os.path.sep) for file in expected]
        if isinstance(files, str):
            files = files.replace("/", os.path.sep)
        elif files is not None:
            files = [file.replace("/", os.path.sep) for file in files]

    # relative path

    audeer.create_archive(
        root,
        files,
        archive_create,
    )
    result = audeer.extract_archive(
        archive_extract,
        root,
        keep_archive=True,
    )
    assert result == expected
    for file in result:
        assert os.path.exists(audeer.path(root, file))
    assert os.path.exists(archive_extract)

    # list of archives

    result = audeer.extract_archives(
        [archive_extract, archive_extract],
        destination,
        keep_archive=True,
    )
    assert result == expected + expected
    for file in result:
        assert os.path.exists(audeer.path(root, file))
    assert os.path.exists(archive_extract)

    # absolute path

    if files is not None:
        if isinstance(files, str):
            files = audeer.path(root, files)
        else:
            files = [audeer.path(root, file) for file in files]
        audeer.create_archive(
            root,
            files,
            archive_create,
        )
        result = audeer.extract_archive(
            archive_extract,
            destination,
            keep_archive=True,
        )
        assert result == expected
        for file in result:
            assert os.path.exists(audeer.path(root, file))
        assert os.path.exists(archive_extract)

    # delete archive

    audeer.extract_archive(
        archive_extract,
        destination,
        keep_archive=False,
    )
    assert not os.path.exists(archive_extract)


@pytest.mark.parametrize(
    "path,ext,basename",
    [
        ("~/.bashrc", None, ".bashrc"),
        ("file.tar.gz", None, "file.tar"),
        ("/a/c.d/g", None, "g"),
        (b"/a/c.d/g", None, "g"),
        ("/a/c.d/g.exe", "exe", "g"),
        ("../.././README.md", ".md", "README"),
        ("folder/file.txt", None, "file"),
        ("folder/file.txt", "txt", "file"),
        (b"folder/file.txt", "txt", "file"),
    ],
)
def test_basename_wo_ext(path, ext, basename):
    b = audeer.basename_wo_ext(path, ext=ext)
    assert b == basename
    assert isinstance(b, str)


@pytest.mark.parametrize(
    "dirs,expected",
    [
        ([], ""),
        (
            [
                "/home/user/tmp/coverage/test",
                "/home/user/tmp/covert/operator",
                "/home/user/tmp/coven/members",
            ],
            "/home/user/tmp",
        ),
        (
            [
                "/home/user/tmp/coverage/test",
                "/home/user/tmp/covert/operator",
                "/home/user/tmp",
            ],
            "/home/user/tmp",
        ),
        (
            [
                "~/tmp/coverage/test",
                "~/tmp/covert/operator",
                "~/tmp/coven/members",
            ],
            f'{os.path.expanduser("~")}/tmp',
        ),
        (
            [
                "/etc/tmp/coverage/test",
                "/home/user/tmp/covert/operator",
                "/home/user/tmp/coven/members",
            ],
            "",
        ),
        (
            [
                "/home/user/tmp",
                "/home/user/tmp",
            ],
            "/home/user/tmp",
        ),
        (
            [
                "/home1/user/tmp",
                "/home2/user/tmp",
            ],
            "",
        ),
    ],
)
def test_common_directory(dirs, expected):
    common = audeer.common_directory(dirs)
    # Change paths always to Linux syntax
    _, common = os.path.splitdrive(common)
    _, expected = os.path.splitdrive(expected)
    common = common.replace("\\", "/")
    expected = expected.replace("\\", "/")
    # On MacOS we get a '/System/Volumes/Data' in front
    common = common.replace("/System/Volumes/Data", "")
    assert common == expected


def test_download_url(tmpdir):
    url = "https://audeering.github.io/audeer/_static/favicon.png"
    audeer.download_url(url, tmpdir)
    audeer.download_url(url, tmpdir)
    dst = audeer.download_url(url, tmpdir, force_download=True)
    assert dst == os.path.join(tmpdir, os.path.basename(url))


@pytest.mark.parametrize(
    "path,extension",
    [
        ("", ""),
        ("~/.bashrc", ""),
        ("file.tar.gz", "gz"),
        ("/a/c.d/g", ""),
        ("/a/c.d/g.exe", "exe"),
        ("../.././README.md", "md"),
        (b"../.././README.md", "md"),
        ("folder/file.txt", "txt"),
        (b"folder/file.txt", "txt"),
        ("test.WAV", "WAV"),
        ("test.WaV", "WaV"),
    ],
)
def test_file_extension(path, extension):
    ext = audeer.file_extension(path)
    assert ext == extension
    assert isinstance(ext, str)


@pytest.mark.parametrize(
    "dir_list,expected,recursive,hidden",
    [
        ([], [], False, False),
        ([], [], True, False),
        (["a", "b", "c"], ["a", "b", "c"], False, False),
        (["a", "b", "c"], ["a", "b", "c"], True, False),
        (["a"], ["a"], False, False),
        (["a"], ["a"], True, False),
        (
            ["a", os.path.join("a", "b"), os.path.join("a", "b", "c")],
            ["a", os.path.join("a", "b"), os.path.join("a", "b", "c")],
            True,
            False,
        ),
        # hidden
        (["a", ".b"], ["a"], True, False),
        (["a", ".b"], [".b", "a"], True, True),
        (
            ["a", ".b", os.path.join("a", ".b"), os.path.join("a", ".b", "c")],
            ["a"],
            True,
            False,
        ),
        (
            ["a", ".b", os.path.join("a", ".b"), os.path.join("a", ".b", "c")],
            [".b", "a", os.path.join("a", ".b"), os.path.join("a", ".b", "c")],
            True,
            True,
        ),
    ],
)
def test_list_dir_names(tmpdir, dir_list, expected, recursive, hidden):
    dir_tmp = tmpdir.mkdir("folder")
    directories = []
    for directory in dir_list:
        directory = os.path.join(str(dir_tmp), directory)
        directories.append(audeer.mkdir(directory))

    for directory in directories:
        assert os.path.isdir(directory)

    path = os.path.join(str(dir_tmp), ".")
    dirs = audeer.list_dir_names(
        path,
        basenames=False,
        recursive=recursive,
        hidden=hidden,
    )
    assert dirs == [audeer.path(dir_tmp, d) for d in expected]
    assert isinstance(dirs, list)

    # test basenames
    dirs = audeer.list_dir_names(
        path,
        basenames=True,
        recursive=recursive,
        hidden=hidden,
    )
    assert dirs == expected


def test_list_dir_names_errors(tmpdir):
    with pytest.raises(NotADirectoryError):
        file = audeer.touch(tmpdir, "file.txt")
        audeer.list_dir_names(file)
    with pytest.raises(FileNotFoundError):
        audeer.list_dir_names("not-existent")


@pytest.mark.parametrize(
    "files,path,filetype,expected,recursive,hidden",
    [
        # empty
        ([], ".", "", [], False, False),
        ([], ".", "", [], True, False),
        ([], ".", "wav", [], False, False),
        ([], ".", "wav", [], True, False),
        # file
        (
            ["file.txt"],
            "file.txt",
            "",
            ["file.txt"],
            False,
            False,
        ),
        pytest.param(
            [os.path.join("sub", "file.txt")],
            "file.txt",
            "",
            [],
            False,
            False,
            marks=pytest.mark.xfail(raises=NotADirectoryError),
        ),
        (
            [os.path.join("sub", "file.txt")],
            "file.txt",
            "",
            [os.path.join("sub", "file.txt")],
            True,
            False,
        ),
        pytest.param(
            [],
            "file",
            "",
            None,
            False,
            False,
            marks=pytest.mark.xfail(raises=NotADirectoryError),
        ),
        (
            [
                "t1.wav",
                "t2.wav",
                os.path.join("sub", "t1.wav"),
                os.path.join("sub", "t2.wav"),
            ],
            "t1.wav",
            "",
            [os.path.join("sub", "t1.wav"), "t1.wav"],
            True,
            False,
        ),
        # folder
        (
            ["t3.ogg", "t2.wav", "t1.wav"],
            ".",
            "",
            ["t1.wav", "t2.wav", "t3.ogg"],
            False,
            False,
        ),
        (
            [
                "t3.ogg",
                "t2.wav",
                "t1.wav",
                os.path.join("sub", "t1.wav"),
                os.path.join("sub", "t1.ogg"),
            ],
            ".",
            "",
            ["t1.wav", "t2.wav", "t3.ogg"],
            False,
            False,
        ),
        (
            [
                "t3.ogg",
                "t2.wav",
                "t1.wav",
                os.path.join("sub", "t1.wav"),
                os.path.join("sub", "sub", "t1.ogg"),
            ],
            ".",
            "",
            [
                os.path.join("sub", "sub", "t1.ogg"),
                os.path.join("sub", "t1.wav"),
                "t1.wav",
                "t2.wav",
                "t3.ogg",
            ],
            True,
            False,
        ),
        pytest.param(
            [],
            "does-not-exist",
            "",
            None,
            False,
            False,
            marks=pytest.mark.xfail(raises=NotADirectoryError),
        ),
        pytest.param(
            [],
            os.path.join("does", "not", "exist"),
            "",
            None,
            False,
            False,
            marks=pytest.mark.xfail(raises=NotADirectoryError),
        ),
        # filetype
        (
            [
                "t3.ogg",
                "t2.wav",
                "t1.wav",
                os.path.join("sub", "t1.wav"),
                os.path.join("sub", "sub", "t1.ogg"),
            ],
            ".",
            "ogg",
            ["t3.ogg"],
            False,
            False,
        ),
        (
            [
                "t3.ogg",
                "t2.wav",
                "t1.wav",
                os.path.join("sub", "t1.wav"),
                os.path.join("sub", "sub", "t1.ogg"),
            ],
            ".",
            "ogg",
            [
                os.path.join("sub", "sub", "t1.ogg"),
                "t3.ogg",
            ],
            True,
            False,
        ),
        (
            [
                "t1.wav",
                os.path.join("sub", "t1.wav"),
            ],
            "t1.wav",
            "",
            ["t1.wav"],
            False,
            False,
        ),
        # pattern
        (
            [
                "t1.wav",
                "t2.ogg",
                "s3.wav",
                os.path.join("sub", "t1.wav"),
                os.path.join("sub", "t2.ogg"),
                os.path.join("sub", "s3.wav"),
            ],
            "t*",
            "",
            [
                os.path.join("sub", "t1.wav"),
                os.path.join("sub", "t2.ogg"),
                "t1.wav",
                "t2.ogg",
            ],
            True,
            False,
        ),
        (
            [
                os.path.join("sub", "t1.wav"),
                os.path.join("sub", "t2.ogg"),
                os.path.join("sub", "s3.wav"),
            ],
            "t*",
            "",
            [
                os.path.join("sub", "t1.wav"),
                os.path.join("sub", "t2.ogg"),
            ],
            True,
            False,
        ),
        (
            [
                "t1.wav",
                "t2.ogg",
                "s3.wav",
                os.path.join("sub", "t1.wav"),
                os.path.join("sub", "t2.ogg"),
                os.path.join("sub", "s3.wav"),
            ],
            "x*",
            "",
            [],
            True,
            False,
        ),
        (
            [
                "t1.wav",
                "t2.wav",
                "s1.wav",
                "s2.wav",
            ],
            "t?.wav",
            "",
            ["t1.wav", "t2.wav"],
            False,
            False,
        ),
        (
            [
                "t1.wav",
                "t2.wav",
                "s1.wav",
                "s2.wav",
            ],
            "[ts]1.wav",
            "",
            ["s1.wav", "t1.wav"],
            False,
            False,
        ),
        (
            [
                "t1.wav",
                "t2.wav",
                "s1.wav",
                "s.wav",
            ],
            "[ts]?.wav",
            "",
            ["s1.wav", "t1.wav", "t2.wav"],
            False,
            False,
        ),
        (
            [
                "t1.wav",
                "t2.wav",
                "s1.wav",
                "s.wav",
                "x.wav",
            ],
            "[ts]*.wav",
            "",
            ["s.wav", "s1.wav", "t1.wav", "t2.wav"],
            False,
            False,
        ),
        (
            [],
            "file*",
            "",
            [],
            False,
            False,
        ),
        (
            [],
            "?ile",
            "",
            [],
            False,
            False,
        ),
        pytest.param(
            [],
            os.path.join("does", "not", "exist", "file*"),
            "",
            None,
            False,
            False,
            marks=pytest.mark.xfail(raises=NotADirectoryError),
        ),
        pytest.param(
            [],
            os.path.join("not!a[pattern"),
            "",
            None,
            False,
            False,
            marks=pytest.mark.xfail(raises=NotADirectoryError),
        ),
        # pattern + filetype
        (
            [
                "t1.wav",
                "t2.ogg",
                "s3.wav",
                os.path.join("sub", "t1.wav"),
                os.path.join("sub", "t2.ogg"),
                os.path.join("sub", "s3.wav"),
            ],
            "t*",
            "ogg",
            [os.path.join("sub", "t2.ogg"), "t2.ogg"],
            True,
            False,
        ),
        # hidden
        (
            [".file.txt"],
            ".file.txt",
            "",
            [],
            False,
            False,
        ),
        (
            [".file.txt"],
            ".file.txt",
            "",
            [".file.txt"],
            False,
            True,
        ),
        (
            [os.path.join("sub", ".file.txt")],
            ".file.txt",
            "",
            [],
            True,
            False,
        ),
        (
            [os.path.join("sub", ".file.txt")],
            ".file.txt",
            "",
            [os.path.join("sub", ".file.txt")],
            True,
            True,
        ),
        (
            [
                "t1.wav",
                ".t2.wav",
                os.path.join("sub", "t1.wav"),
                os.path.join("sub", ".t2.wav"),
            ],
            "",
            "",
            [
                "t1.wav",
            ],
            False,
            False,
        ),
        (
            [
                "t1.wav",
                ".t2.wav",
                os.path.join("sub", "t1.wav"),
                os.path.join("sub", ".t2.wav"),
            ],
            "",
            "",
            [
                os.path.join("sub", "t1.wav"),
                "t1.wav",
            ],
            True,
            False,
        ),
        (
            [
                "t1.wav",
                ".t2.wav",
                os.path.join("sub", "t1.wav"),
                os.path.join("sub", ".t2.wav"),
            ],
            "",
            "",
            [
                ".t2.wav",
                "t1.wav",
            ],
            False,
            True,
        ),
        (
            [
                "t1.wav",
                ".t2.wav",
                os.path.join("sub", "t1.wav"),
                os.path.join("sub", ".t2.wav"),
            ],
            "",
            "",
            [
                ".t2.wav",
                os.path.join("sub", ".t2.wav"),
                os.path.join("sub", "t1.wav"),
                "t1.wav",
            ],
            True,
            True,
        ),
    ],
)
def test_list_file_names(tmpdir, files, path, filetype, expected, recursive, hidden):
    dir_tmp = tmpdir.mkdir("folder")
    dir_tmp.mkdir("subfolder")
    path = os.path.join(str(dir_tmp), path)
    for file in files:
        # Create the files
        file_tmp = dir_tmp.join(file)
        audeer.mkdir(os.path.dirname(file_tmp))
        file_tmp.write("")
    f = audeer.list_file_names(
        path,
        filetype=filetype,
        basenames=False,
        recursive=recursive,
        hidden=hidden,
    )
    # test full path
    assert f == [audeer.path(dir_tmp, f) for f in expected]
    assert isinstance(f, list)
    # test basenames
    f = audeer.list_file_names(
        path,
        filetype=filetype,
        basenames=True,
        recursive=recursive,
        hidden=hidden,
    )
    assert f == expected


def test_list_file_names_symlinks(tmpdir):
    # symlinks in folder
    folder = audeer.mkdir(tmpdir, "folder")
    file = audeer.touch(folder, "file.txt")
    link = audeer.path(folder, "link.txt")
    os.symlink(file, link)
    files = audeer.list_file_names(folder, basenames=True)
    assert files == ["file.txt", "link.txt"]
    os.remove(link)
    # symlinks to sub-folders
    sub_folder = audeer.mkdir(folder, "sub-folder")
    audeer.touch(sub_folder, "file2.txt")
    link = audeer.path(folder, "link")
    os.symlink(sub_folder, link)
    files = audeer.list_file_names(folder, basenames=True)
    assert files == ["file.txt"]
    files = audeer.list_file_names(folder, basenames=True, recursive=True)
    assert files == [
        "file.txt",
        os.path.join("link", "file2.txt"),
        os.path.join("sub-folder", "file2.txt"),
    ]


def test_md5_errors():
    with pytest.raises(FileNotFoundError):
        audeer.md5("does/not/exist")


@pytest.mark.parametrize(
    "file, content, expected",
    [
        (  # empty file
            "file.txt",
            None,
            "d41d8cd98f00b204e9800998ecf8427e",
        ),
        (  # different content
            "file.txt",
            "hello world",
            "5eb63bbbe01eeed093cb22bb8f5acdc3",
        ),
        (
            "file.txt",
            "Hello World",
            "b10a8db164e0754105b7a99be72e3fe5",
        ),
        (  # different filename
            "file.TXT",
            "Hello World",
            "b10a8db164e0754105b7a99be72e3fe5",
        ),
    ],
)
def test_md5_file(tmpdir, file, content, expected):
    path = audeer.path(tmpdir, file)
    with open(path, "w") as fp:
        if content is not None:
            fp.write(content)

    assert audeer.md5(path) == expected


@pytest.mark.parametrize(
    "tree, content, expected",
    [
        (  # empty folder
            [],
            None,
            "d41d8cd98f00b204e9800998ecf8427e",
        ),
        (  # folder with different content
            ["f"],
            None,
            "8fa14cdd754f91cc6554c9e71929cce7",
        ),
        (
            ["sub/f"],
            None,
            "1af042d5a4ec129583f6093f98f64118",
        ),
        (
            ["f", "sub/f"],
            None,
            "b540f38948f445622adc657a757f4b0d",
        ),
        (
            ["f", "sub/g"],
            None,
            "305107efbb15f9334d22ae4fbeec4de6",
        ),
        (
            ["f", "sub/g"],
            "hello world",
            "47829eb8ef287d0d72e0fed9b96d258d",
        ),
        (
            ["f", "sub/g"],
            "Hello World",
            "442d96d7c43bb18f247888408e5d6977",
        ),
        (  # with empty sub folder
            ["f", "sub/g", "sub/"],
            None,
            "305107efbb15f9334d22ae4fbeec4de6",
        ),
        (  # with hidden file
            ["f", "sub/g", ".hidden"],
            None,
            "97490b233a7717aec19023e28443a1bf",
        ),
        (  # umlaute
            ["ä", "ö", "ü", "ß"],
            None,
            "622165ad36122984c6b2c7ba466aa262",
        ),
    ],
    indirect=["tree"],
)
def test_md5_folder(tmpdir, tree, content, expected):
    if content is not None:
        for path in tree:
            with open(path, "w") as fp:
                fp.write(content)

    assert audeer.md5(tmpdir) == expected


def test_md5_symbolic_link(tmpdir):
    # Link to a file
    file = audeer.touch(tmpdir, "file.txt")
    link = audeer.path(tmpdir, "link.txt")
    os.symlink(file, link)
    assert audeer.md5(file) == audeer.md5(link)
    os.remove(file)
    os.remove(link)
    # Link to a folder
    folder = audeer.mkdir(tmpdir, "folder")
    file = audeer.touch(folder, "file.txt")
    link = audeer.path(tmpdir, "link")
    os.symlink(folder, link)
    assert audeer.md5(folder) == audeer.md5(link)
    os.remove(link)
    # Link to file in folder
    md5_single_file = audeer.md5(folder)
    link = audeer.path(folder, "link.txt")
    os.symlink(file, link)
    md5_link = audeer.md5(folder)
    os.remove(link)
    audeer.touch(folder, "link.txt")
    md5_file = audeer.md5(folder)
    assert md5_link == md5_file
    assert md5_file != md5_single_file


def test_mkdir(tmpdir):
    # New dir
    path = str(tmpdir.mkdir("folder1"))
    p = audeer.mkdir(path)
    assert os.path.isdir(p) is True
    assert p == path
    # Existing dir
    p = audeer.mkdir(path)
    assert os.path.isdir(p) is True
    assert p == path
    # Existing dir with content
    dir_tmp = tmpdir.mkdir("folder2")
    f = dir_tmp.join("file.txt")
    f.write("")
    path = str(dir_tmp)
    p = audeer.mkdir(path)
    assert os.path.isdir(p) is True
    assert p == path
    # Relative path
    path = str(tmpdir.mkdir("folder3"))
    current_path = os.getcwd()
    os.chdir(path)
    p = audeer.mkdir("folder4")
    os.chdir(current_path)
    assert os.path.isdir(p) is True
    assert p == os.path.join(path, "folder4")
    # Subdirectories
    os.chdir(path)
    p = audeer.mkdir("folder5", "folder6")
    os.chdir(current_path)
    assert os.path.isdir(p) is True
    assert p == os.path.join(path, "folder5", "folder6")
    # Path in bytes
    path = str(tmpdir.mkdir("folder7"))
    path = bytes(path, "utf8")
    p = audeer.mkdir(path)
    assert os.path.isdir(p) is True
    assert p == path.decode("utf8")
    # Empty dir
    path = ""
    p = audeer.mkdir(path)
    assert p == path
    # Mode, see https://stackoverflow.com/a/705088
    # Default mode
    os.umask(0)
    p = audeer.mkdir(tmpdir, "folder8", "sub-folder")
    mode = stat.S_IMODE(os.stat(p).st_mode)
    expected_mode = int("777", 8)
    assert mode == expected_mode
    # Non-default modes
    # Under Windows, changing permissions does not work,
    # there we always expect 777
    os.umask(0)
    p = audeer.mkdir(tmpdir, "folder9", "sub-folder", mode=0o775)
    expected_mode = "775"
    if platform.system() == "Windows":
        expected_mode = "777"
    mode = stat.S_IMODE(os.stat(p).st_mode)
    assert mode == int(expected_mode, 8)
    assert mode != int("755", 8)
    os.umask(0)
    p = audeer.mkdir(tmpdir, "folder10", "sub-folder", mode=0o755)
    expected_mode = "755"
    if platform.system() == "Windows":
        expected_mode = "777"
    mode = stat.S_IMODE(os.stat(p).st_mode)
    assert mode == int(expected_mode, 8)
    assert mode != int("775", 8)


@pytest.mark.parametrize(
    "src_path, dst_path",
    [
        (
            "path1",
            "path1",
        ),
        (
            "path1",
            "path2",
        ),
    ],
)
def test_move(tmpdir, src_path, dst_path):
    system = platform.system()
    tmp_dir = audeer.mkdir(tmpdir, "folder")

    # src: file
    # dst: new file
    audeer.touch(tmp_dir, src_path)
    audeer.move(
        os.path.join(tmp_dir, src_path),
        os.path.join(tmp_dir, dst_path),
    )
    if src_path != dst_path:
        assert not os.path.exists(os.path.join(tmp_dir, src_path))
    assert os.path.exists(os.path.join(tmp_dir, dst_path))

    # src: file
    # dst: existing file
    audeer.rmdir(tmp_dir)
    audeer.mkdir(tmp_dir)
    audeer.touch(tmp_dir, src_path)
    audeer.touch(tmp_dir, dst_path)
    audeer.move(
        os.path.join(tmp_dir, src_path),
        os.path.join(tmp_dir, dst_path),
    )
    if src_path != dst_path:
        assert not os.path.exists(os.path.join(tmp_dir, src_path))
    assert os.path.exists(os.path.join(tmp_dir, dst_path))

    # src: folder
    # dst: new folder
    audeer.rmdir(tmp_dir)
    audeer.mkdir(tmp_dir, src_path)
    audeer.touch(tmp_dir, src_path, "file.txt")
    audeer.move(
        os.path.join(tmp_dir, src_path),
        os.path.join(tmp_dir, dst_path),
    )
    if src_path != dst_path:
        assert not os.path.exists(os.path.join(tmp_dir, src_path))
    assert os.path.exists(os.path.join(tmp_dir, dst_path))
    assert os.path.exists(os.path.join(tmp_dir, dst_path, "file.txt"))
    assert os.path.isdir(os.path.join(tmp_dir, dst_path))

    # src: non-empty folder
    # dst: existing non-empty folder
    audeer.rmdir(tmp_dir)
    audeer.mkdir(tmp_dir, src_path)
    audeer.touch(tmp_dir, src_path, "file.txt")
    if src_path != dst_path:
        audeer.mkdir(tmp_dir, dst_path)
        audeer.touch(tmp_dir, dst_path, "file.txt")
    if src_path != dst_path:
        if system == "Windows":
            error_msg = "Access is denied"
        else:
            error_msg = "Directory not empty"
        with pytest.raises(OSError, match=error_msg):
            audeer.move(
                os.path.join(tmp_dir, src_path),
                os.path.join(tmp_dir, dst_path),
            )

        # src: non-empty folder
        # dst: existing empty folder
        os.remove(os.path.join(tmp_dir, dst_path, "file.txt"))
        if system == "Windows":
            # Only under Windows
            # we get an error
            # if destination is an empty folder
            with pytest.raises(OSError, match="Access is denied"):
                audeer.move(
                    os.path.join(tmp_dir, src_path),
                    os.path.join(tmp_dir, dst_path),
                )
        else:
            audeer.move(
                os.path.join(tmp_dir, src_path),
                os.path.join(tmp_dir, dst_path),
            )

            if src_path != dst_path:
                assert not os.path.exists(os.path.join(tmp_dir, src_path))
            assert os.path.exists(os.path.join(tmp_dir, dst_path))
            assert os.path.exists(os.path.join(tmp_dir, dst_path, "file.txt"))
            assert os.path.isdir(os.path.join(tmp_dir, dst_path))

    # src: non-empty folder
    # dst: identical to src
    if src_path == dst_path:
        audeer.move(
            os.path.join(tmp_dir, src_path),
            os.path.join(tmp_dir, dst_path),
        )

        if src_path != dst_path:
            assert not os.path.exists(os.path.join(tmp_dir, src_path))
        assert os.path.exists(os.path.join(tmp_dir, dst_path))
        assert os.path.exists(os.path.join(tmp_dir, dst_path, "file.txt"))
        assert os.path.isdir(os.path.join(tmp_dir, dst_path))

    # src: empty folder
    # dst: existing non-empty folder
    audeer.rmdir(tmp_dir)
    audeer.mkdir(tmp_dir, src_path)
    if src_path != dst_path:
        audeer.mkdir(tmp_dir, dst_path)
        audeer.touch(tmp_dir, dst_path, "file.txt")
    if src_path != dst_path:
        if system == "Windows":
            error_msg = "Access is denied"
        else:
            error_msg = "Directory not empty"
        with pytest.raises(OSError, match=error_msg):
            audeer.move(
                os.path.join(tmp_dir, src_path),
                os.path.join(tmp_dir, dst_path),
            )

        # src: empty folder
        # dst: existing empty folder
        os.remove(os.path.join(tmp_dir, dst_path, "file.txt"))
        if system == "Windows":
            # Only under Windows
            # we get an error
            # if destination is an empty folder
            with pytest.raises(OSError, match="Access is denied"):
                audeer.move(
                    os.path.join(tmp_dir, src_path),
                    os.path.join(tmp_dir, dst_path),
                )
        else:
            audeer.move(
                os.path.join(tmp_dir, src_path),
                os.path.join(tmp_dir, dst_path),
            )

            if src_path != dst_path:
                assert not os.path.exists(os.path.join(tmp_dir, src_path))
            assert os.path.exists(os.path.join(tmp_dir, dst_path))
            assert not os.path.exists(os.path.join(tmp_dir, dst_path, "file.txt"))
            assert os.path.isdir(os.path.join(tmp_dir, dst_path))

    # src: empty folder
    # dst: identical to src
    if src_path == dst_path:
        audeer.move(
            os.path.join(tmp_dir, src_path),
            os.path.join(tmp_dir, dst_path),
        )

        if src_path != dst_path:
            assert not os.path.exists(os.path.join(tmp_dir, src_path))
        assert os.path.exists(os.path.join(tmp_dir, dst_path))
        assert not os.path.exists(os.path.join(tmp_dir, dst_path, "file.txt"))
        assert os.path.isdir(os.path.join(tmp_dir, dst_path))

    if src_path != dst_path:
        # src: file
        # dst: non-empty folder
        audeer.rmdir(tmp_dir)
        audeer.mkdir(tmp_dir)
        audeer.touch(tmp_dir, src_path)
        audeer.mkdir(tmp_dir, dst_path)
        audeer.touch(tmp_dir, dst_path, "file.txt")
        if system == "Windows":
            error_msg = "Access is denied"
        else:
            error_msg = "Is a directory"
        with pytest.raises(OSError, match=error_msg):
            audeer.move(
                os.path.join(tmp_dir, src_path),
                os.path.join(tmp_dir, dst_path),
            )

        # src: file
        # dst: empty folder
        os.remove(os.path.join(tmp_dir, dst_path, "file.txt"))
        with pytest.raises(OSError, match=error_msg):
            audeer.move(
                os.path.join(tmp_dir, src_path),
                os.path.join(tmp_dir, dst_path),
            )

        # src: non-empty folder
        # dst: file
        audeer.rmdir(tmp_dir)
        audeer.mkdir(tmp_dir)
        audeer.mkdir(tmp_dir, src_path)
        audeer.touch(tmp_dir, src_path, "file.txt")
        audeer.touch(tmp_dir, dst_path)
        if system != "Windows":
            error_msg = "Not a directory"
            with pytest.raises(OSError, match=error_msg):
                audeer.move(
                    os.path.join(tmp_dir, src_path),
                    os.path.join(tmp_dir, dst_path),
                )
        else:
            audeer.move(
                os.path.join(tmp_dir, src_path),
                os.path.join(tmp_dir, dst_path),
            )
            assert not os.path.exists(os.path.join(tmp_dir, src_path))
            assert os.path.exists(os.path.join(tmp_dir, dst_path))
            assert os.path.exists(os.path.join(tmp_dir, dst_path, "file.txt"))
            assert os.path.isdir(os.path.join(tmp_dir, dst_path))

        # src: empty folder
        # dst: file
        audeer.rmdir(tmp_dir)
        audeer.mkdir(tmp_dir, src_path)
        audeer.touch(tmp_dir, dst_path)
        if system != "Windows":
            error_msg = "Not a directory"
            with pytest.raises(OSError, match=error_msg):
                audeer.move(
                    os.path.join(tmp_dir, src_path),
                    os.path.join(tmp_dir, dst_path),
                )
        else:
            audeer.move(
                os.path.join(tmp_dir, src_path),
                os.path.join(tmp_dir, dst_path),
            )
            assert not os.path.exists(os.path.join(tmp_dir, src_path))
            assert os.path.exists(os.path.join(tmp_dir, dst_path))
            assert not os.path.exists(os.path.join(tmp_dir, dst_path, "file.txt"))
            assert os.path.isdir(os.path.join(tmp_dir, dst_path))


@pytest.mark.parametrize(
    "src_file, dst_file",
    [
        (
            "file1",
            "file1",
        ),
        (
            "file1",
            "file2",
        ),
    ],
)
def test_move_file(tmpdir, src_file, dst_file):
    tmp_path = audeer.mkdir(tmpdir, "folder")

    src_path = audeer.touch(tmp_path, src_file)
    dst_path = os.path.join(tmp_path, dst_file)

    audeer.move_file(src_path, dst_path)

    if src_file != dst_file:
        assert not os.path.exists(src_path)
    assert os.path.exists(dst_path)


@pytest.mark.parametrize(
    "path, new_extension, ext, expected_path",
    [
        ("", "", None, ""),
        ("", "txt", None, ""),
        ("", "", "rst", ""),
        ("", "txt", "rst", ""),
        ("file", "", None, "file"),
        ("file", "txt", None, "file.txt"),
        ("file.txt", "wav", None, "file.wav"),
        ("test/file.txt", "wav", None, "test/file.wav"),
        ("a/b/../file.txt", "wav", None, "a/b/../file.wav"),
        ("file.txt", "wav", "txt", "file.wav"),
        ("file.txt", "wav", ".txt", "file.wav"),
        ("file.txt", ".wav", "txt", "file.wav"),
        ("file.txt", ".wav", ".txt", "file.wav"),
        ("file.a.b", "wav", "a.b", "file.wav"),
        ("file.a.b", "wav", ".a.b", "file.wav"),
        ("file", "wav", "ext", "file"),
        ("file.txt", "wav", "ext", "file.txt"),
        ("file.txt", "wav", "t", "file.txt"),
    ],
)
def test_replace_file_extension(path, new_extension, ext, expected_path):
    new_path = audeer.replace_file_extension(path, new_extension, ext=ext)
    assert new_path == expected_path


def test_rmdir(tmpdir):
    # Non existing dir
    audeer.rmdir("non-esitent")
    # Folder with file content
    dir_tmp = tmpdir.mkdir("folder")
    f = dir_tmp.join("file.txt")
    f.write("")
    path = str(dir_tmp)
    p = audeer.mkdir(path)
    with pytest.raises(NotADirectoryError):
        audeer.rmdir(os.path.join(p, "file.txt"))
    audeer.rmdir(p)
    assert not os.path.exists(p)
    # Folder with folder content
    p = audeer.mkdir(tmpdir, "folder", "sub-folder")
    audeer.rmdir(tmpdir, "folder")
    assert not os.path.exists(p)
    assert not os.path.exists(os.path.dirname(p))
    # Relative path
    path = str(tmpdir.mkdir("folder"))
    current_path = os.getcwd()
    os.chdir(os.path.dirname(path))
    assert os.path.exists(path)
    audeer.rmdir("folder")
    assert not os.path.exists(path)
    os.chdir(current_path)
    # Symbolic link
    path = audeer.mkdir(tmpdir, "folder")
    link = os.path.join(tmpdir, "link")
    os.symlink(path, link)
    # Error message is broken
    # for newer version of Python 3.12
    # under MacOS and Linux
    python_version = (
        sys.version_info.major,
        sys.version_info.minor,
        sys.version_info.micro,
    )
    if python_version >= (3, 12, 4) and platform.system() != "Windows":
        error_msg = "None"
    else:
        error_msg = "symbolic link"
    with pytest.raises(OSError, match=error_msg):
        audeer.rmdir(link, follow_symlink=False)
    assert os.path.exists(link)
    assert os.path.exists(path)
    audeer.rmdir(link, follow_symlink=True)
    assert not os.path.exists(link)
    assert not os.path.exists(path)


def test_script_dir(tmpdir):
    r"""Test estimation of current directory of caller.

    See https://stackoverflow.com/a/5137509.

    Args:
        tmpdir: tmpdir fixture

    """
    expected_script_dir = os.path.dirname(os.path.realpath(__file__))
    assert audeer.script_dir() == expected_script_dir
    current_dir = os.getcwd()
    os.chdir(tmpdir)
    assert audeer.script_dir() == expected_script_dir
    os.chdir(current_dir)


def test_touch(tmpdir):
    path = audeer.mkdir(tmpdir, "folder1")
    path = os.path.join(path, "file")
    assert not os.path.exists(path)
    audeer.touch(path)
    assert os.path.exists(path)
    stat = os.stat(path)
    time.sleep(0.1)
    audeer.touch(path)
    assert os.path.exists(path)
    new_stat = os.stat(path)
    assert stat.st_atime < new_stat.st_atime
    assert stat.st_mtime < new_stat.st_mtime



import importlib
import subprocess
import sys

import pytest

import audeer


PACKAGE = "PyYaml"
MODULE = "yaml"


def uninstall():
    subprocess.run(
        [
            sys.executable,
            "-m",
            "pip",
            "uninstall",
            "--yes",
            PACKAGE,
        ],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    # remove from module cache
    remove = []
    for module in sys.modules:
        if module.startswith(MODULE):
            remove.append(module)
    for module in remove:
        sys.modules.pop(module)


@pytest.fixture(autouse=True)
def run_around_tests():
    yield
    uninstall()


def test():
    # verify module is not installed
    with pytest.raises(ModuleNotFoundError):
        importlib.import_module(MODULE)

    # version not found
    with pytest.raises(subprocess.CalledProcessError):
        audeer.install_package(
            PACKAGE,
            version="999.0.0",
        )

    # install package
    audeer.install_package(
        PACKAGE,
        version="<=5.3",
    )

    # installed version satisfies requested version
    audeer.install_package(
        PACKAGE,
        version=">=5.3",
    )
    audeer.install_package(
        PACKAGE,
        version=">5.2",
    )
    audeer.install_package(
        PACKAGE,
        version="<=6.0",
    )
    audeer.install_package(
        PACKAGE,
        version="  <   5.4  ",  # whitespace will be ignored
    )
    audeer.install_package(
        PACKAGE,
        version=None,
    )

    # installed version does not satisfied requested version
    with pytest.raises(RuntimeError):
        audeer.install_package(
            PACKAGE,
            version=">=5.4",
        )
    with pytest.raises(RuntimeError):
        audeer.install_package(
            PACKAGE,
            version=">5.3",
        )
    with pytest.raises(RuntimeError):
        audeer.install_package(
            PACKAGE,
            version="<=5.2",
        )
    with pytest.raises(RuntimeError):
        audeer.install_package(
            PACKAGE,
            version="<5.3",
        )



import time

import pytest

import audeer


@pytest.mark.parametrize(
    "text",
    [
        ("Hello world"),
        (4 * "abcdefghijklmnopqrstuvwxyz"),
        (18 * "a"),  # 18 == audeer.config.TQDM_COLUMNS as returned by doctest
    ],
)
def test_format_display_message(text):
    t = audeer.format_display_message(text, pbar=False)
    assert len(t) == audeer.config.TQDM_COLUMNS - 2
    if len(text) > audeer.config.TQDM_COLUMNS:
        m = (audeer.config.TQDM_COLUMNS - 3) // 2
        assert t[m - 1 : m + 2] == "..."
    assert t.startswith(text[:2])
    t = audeer.format_display_message(text, pbar=True)
    assert len(t) == audeer.config.TQDM_DESCLEN - 2
    if len(text) > audeer.config.TQDM_DESCLEN:
        m = (audeer.config.TQDM_DESCLEN - 3) // 2
        assert t[m - 1 : m + 2] == "..."
    assert t.startswith(text[:2])


def test_progress_bar():
    assert audeer.config.TQDM_DESCLEN == 60
    assert audeer.config.TQDM_FORMAT == (
        "{percentage:3.0f}%|{bar} [{elapsed}<{remaining}] "
        "{desc:" + str(audeer.config.TQDM_DESCLEN) + "}"
    )
    pbar = audeer.progress_bar([0.1])
    for step in pbar:
        time.sleep(step)


def test_progress_bar_update():
    r"""Ensure progress bar is refreshed.

    If the progress bar has to wait for a long time
    until it would get updated,
    we enforce an update by a given time.

    """
    for _ in audeer.progress_bar(range(2), maximum_refresh_time=0.01):
        time.sleep(0.05)



# This code comes from
# https://github.com/python/cpython/blob/20a1c8ee4bcb1c421b7cca1f3f5d6ad7ce30a9c9/Lib/distutils/tests/test_version.py
# and is licensed under
# https://github.com/python/cpython/blob/20a1c8ee4bcb1c421b7cca1f3f5d6ad7ce30a9c9/LICENSE

import re


class Version:
    """Abstract base class for version numbering classes."""

    def __repr__(self):
        r"""Python code to recreate instance."""
        return f"{self.__class__.__name__} ('{str(self)}')"

    def __eq__(self, other):
        r"""Check if version equals another version."""
        c = self._cmp(other)
        return c == 0

    def __lt__(self, other):
        r"""Check if version is less than another version."""
        c = self._cmp(other)
        return c < 0

    def __le__(self, other):
        r"""Check if version is less or equal another version."""
        c = self._cmp(other)
        return c <= 0

    def __gt__(self, other):
        r"""Check if version is greater than another version."""
        c = self._cmp(other)
        return c > 0

    def __ge__(self, other):
        r"""Check if version is greater or equal another version."""
        c = self._cmp(other)
        return c >= 0


# Interface for version-number classes -- must be implemented
# by the following classes (the concrete ones -- Version should
# be treated as an abstract class).
#    __init__ (string) - create and take same action as 'parse'
#                        (string parameter is optional)
#    parse (string)    - convert a string representation to whatever
#                        internal representation is appropriate for
#                        this style of version numbering
#    __str__ (self)    - convert back to a string; should be very similar
#                        (if not identical to) the string supplied to parse
#    __repr__ (self)   - generate Python code to recreate
#                        the instance
#    _cmp (self, other) - compare two version numbers ('other' may
#                        be an unparsed version string, or another
#                        instance of your version class)


class StrictVersion(Version):
    """Version numbering for anal retentives and software idealists.

    This implementation was originally part of :mod:`distutils`
    as ``version.StrictVersion``.

    A version number consists of two
    or three dot-separated numeric components,
    with an optional "pre-release" tag on the end.
    The pre-release tag consists of the letter 'a' or 'b'
    followed by a number.
    If the numeric components of two version numbers are equal,
    then one with a pre-release tag
    will always be deemed earlier (lesser)
    than one without.

    The following are valid version numbers
    (shown in the order
    that would be obtained
    by sorting according to the supplied cmp function):
    ``'0.4'``,
    ``'0.4.1'``,
    ``'0.5a1'``,
    ``'0.5b3'``,
    ``'0.5'``,
    ``'0.9.6'``,
    ``'1.0'``,
    ``'1.0.4a3'``,
    ``'1.0.4b1'``,
    ``'1.0.4'``.

    The following are examples of invalid version numbers:
    ``'1'``,
    ``'2.7.2.2'``,
    ``'1.3.a4'``,
    ``'1.3pl1'``,
    ``'1.3c4'``.

    Args:
        version: version string

    Raises:
        ValueError: if ``version`` does not match
            the ``StrictVersion.version_re`` pattern

    Examples:
        >>> v1 = audeer.StrictVersion("1.17.2a1")
        >>> v1
        StrictVersion ('1.17.2a1')
        >>> v1.version
        (1, 17, 2)
        >>> v1.prerelease
        ('a', 1)
        >>> v2 = audeer.StrictVersion("1.17.2")
        >>> v1 < v2
        True

    """

    version_re = re.compile(
        r"^(\d+) \. (\d+) (\. (\d+))? ([ab](\d+))?$", re.VERBOSE | re.ASCII
    )
    """Version regexp pattern.

    The regexp pattern is used to split the version
    into
    *major*,
    *minor*,
    *patch*,
    *prerelease*,
    *prerelease number*
    when parsing it.

    """

    def __init__(self, version=None):
        self.version = None
        r"Parsed version."
        self.prerelease = None
        r"Parsed prerelease part of version."

        if version:
            self.parse(version)

    def parse(self, version):
        r"""Parse a version string.

        When called this updates the stored version
        under ``self.version``.

        Args:
            version: version string

        """
        match = self.version_re.match(version)
        if not match:
            raise ValueError(f"invalid version number '{version}'")

        (major, minor, patch, prerelease, prerelease_num) = match.group(1, 2, 4, 5, 6)

        if patch:
            self.version = tuple(map(int, [major, minor, patch]))
        else:
            self.version = tuple(map(int, [major, minor])) + (0,)

        if prerelease:
            self.prerelease = (prerelease[0], int(prerelease_num))
        else:
            self.prerelease = None

    def __str__(self):
        r"""String representation of version."""
        if self.version[2] == 0:
            vstring = ".".join(map(str, self.version[0:2]))
        else:
            vstring = ".".join(map(str, self.version))

        if self.prerelease:
            vstring = vstring + self.prerelease[0] + str(self.prerelease[1])

        return vstring

    def _cmp(self, other):
        if isinstance(other, str):
            other = StrictVersion(other)
        elif not isinstance(other, StrictVersion):
            return NotImplemented

        if self.version != other.version:
            # numeric versions don't match
            # prerelease stuff doesn't matter
            if self.version < other.version:
                return -1
            else:
                return 1

        # have to compare prerelease
        # case 1: neither has prerelease; they're equal
        # case 2: self has prerelease, other doesn't; other is greater
        # case 3: self doesn't have prerelease, other does: self is greater
        # case 4: both have prerelease: must compare them!

        if not self.prerelease and not other.prerelease:
            return 0
        elif self.prerelease and not other.prerelease:
            return -1
        elif not self.prerelease and other.prerelease:
            return 1
        elif self.prerelease and other.prerelease:
            if self.prerelease == other.prerelease:
                return 0
            elif self.prerelease < other.prerelease:
                return -1
            else:
                return 1


# The rules according to Greg Stein:
# 1) a version number has 1 or more numbers separated by a period or by
#    sequences of letters. If only periods, then these are compared
#    left-to-right to determine an ordering.
# 2) sequences of letters are part of the tuple for comparison and are
#    compared lexicographically
# 3) recognize the numeric components may have leading zeroes
#
# The LooseVersion class below implements these rules: a version number
# string is split up into a tuple of integer and string components, and
# comparison is a simple tuple comparison.  This means that version
# numbers behave in a predictable and obvious way, but a way that might
# not necessarily be how people *want* version numbers to behave.  There
# wouldn't be a problem if people could stick to purely numeric version
# numbers: just split on period and compare the numbers as tuples.
# However, people insist on putting letters into their version numbers;
# the most common purpose seems to be:
#   - indicating a "pre-release" version
#     ('alpha', 'beta', 'a', 'b', 'pre', 'p')
#   - indicating a post-release patch ('p', 'pl', 'patch')
# but of course this can't cover all version number schemes, and there's
# no way to know what a programmer means without asking him.
#
# The problem is what to do with letters (and other non-numeric
# characters) in a version number.  The current implementation does the
# obvious and predictable thing: keep them as strings and compare
# lexically within a tuple comparison.  This has the desired effect if
# an appended letter sequence implies something "post-release":
# eg. "0.99" < "0.99pl14" < "1.0", and "5.001" < "5.001m" < "5.002".
#
# However, if letters in a version number imply a pre-release version,
# the "obvious" thing isn't correct.  Eg. you would expect that
# "1.5.1" < "1.5.2a2" < "1.5.2", but under the tuple/lexical comparison
# implemented here, this just isn't so.


class LooseVersion(Version):
    r"""Version numbering for anarchists and software realists.

    This implementation was originally part of :mod:`distutils`
    as ``version.LooseVersion``.

    A version number consists of a series of numbers,
    separated by either periods or strings of letters.
    When comparing version numbers,
    the numeric components will be compared numerically,
    and the alphabetic components lexically.
    The following are all valid version numbers,
    in no particular order:
    ``'1.5.1``,
    ``'1.5.2b2``,
    ``'161'``,
    ``'3.10a'``,
    ``'8.02'``,
    ``'3.4j'``,
    ``'1996.07.12'``,
    ``'3.2.pl0'``,
    ``'3.1.1.6'``,
    ``'2g6'``,
    ``'11g'``,
    ``'0.960923'``,
    ``'2.2beta29'``,
    ``'1.13++'``,
    ``'5.5.kw'``,
    ``'2.0b1pl0'``.

    Args:
        version: version string

    Examples:
        >>> v1 = audeer.LooseVersion("1.17.2")
        >>> v1
        LooseVersion ('1.17.2')
        >>> v1.version
        [1, 17, 2]
        >>> v2 = audeer.LooseVersion("1.17.2-3-g70b71bd")
        >>> v1 < v2
        True

    """

    version_re = re.compile(r"(\d+ | [a-z]+ | \.)", re.VERBOSE)
    """Version regexp pattern.

    The regexp pattern is used to split the version
    into single components
    when parsing it.

    """

    def __init__(self, version=None):
        self.version = None
        r"Parsed version."

        if version:
            self.parse(version)

    def __eq__(self, other):  # noqa: D105 (inherited)
        c = self._cmp(other)
        return c == 0

    def __lt__(self, other):  # noqa: D105 (inherited)
        c = self._cmp(other)
        return c < 0

    def __le__(self, other):  # noqa: D105 (inherited)
        c = self._cmp(other)
        return c <= 0

    def __gt__(self, other):  # noqa: D105 (inherited)
        c = self._cmp(other)
        return c > 0

    def __ge__(self, other):  # noqa: D105 (inherited)
        c = self._cmp(other)
        return c >= 0

    def __str__(self):
        r"""String representation of version."""
        return self.vstring

    def parse(self, version):
        r"""Parse a version string.

        When called this updates the stored version
        under ``self.version``.

        Args:
            version: version string

        """
        # I've given up on thinking I can reconstruct the version string
        # from the parsed tuple -- so I just store the string here for
        # use by __str__
        self.vstring = version
        components = [x for x in self.version_re.split(version) if x and x != "."]
        for i, obj in enumerate(components):
            try:
                components[i] = int(obj)
            except ValueError:
                pass

        self.version = components

    def _cmp(self, other):
        if isinstance(other, str):
            other = LooseVersion(other)
        elif not isinstance(other, LooseVersion):
            return NotImplemented

        if self.version == other.version:
            return 0
        if self.version < other.version:
            return -1
        if self.version > other.version:
            return 1



import operator

import pytest

import audeer


@pytest.mark.parametrize(
    "v1, v2, wanted",
    [
        ("1.5.1", "1.5.2b2", -1),
        ("161", "3.10a", 1),
        ("8.02", "8.02", 0),
        ("3.4j", "1996.07.12", -1),
        ("3.2.pl0", "3.1.1.6", 1),
        ("2g6", "11g", -1),
        ("0.960923", "2.2beta29", -1),
        ("1.13++", "5.5.kw", -1),
    ],
)
def test_loose_version(v1, v2, wanted):
    # This test is replicating the original tests from
    # https://github.com/python/cpython/blob/20a1c8ee4bcb1c421b7cca1f3f5d6ad7ce30a9c9/Lib/distutils/tests/test_version.py

    v1 = audeer.LooseVersion(v1)

    result = v1._cmp(audeer.LooseVersion(v2))
    assert result == wanted

    result = v1._cmp(v2)
    assert result == wanted

    result = v1._cmp(object())
    assert result == NotImplemented

    assert v1.__repr__() == f"LooseVersion ('{str(v1)}')"


@pytest.mark.parametrize(
    "v1, v2, operator, expected",
    [
        ("1.5.1", "1.5.2b2", operator.eq, False),
        ("1.5.1", "1.5.2b2", operator.lt, True),
        ("1.5.1", "1.5.2b2", operator.le, True),
        ("1.5.1", "1.5.2b2", operator.gt, False),
        ("1.5.1", "1.5.2b2", operator.ge, False),
        ("1.5.1", "1.5.1", operator.eq, True),
        ("1.5.1", "1.5.1", operator.le, True),
        ("1.5.1", "1.5.1", operator.ge, True),
        ("161", "3.10a", operator.gt, True),
        ("8.02", "8.02", operator.eq, True),
        ("3.4j", "1996.07.12", operator.lt, True),
        ("3.2.pl0", "3.1.1.6", operator.gt, True),
        ("2g6", "11g", operator.lt, True),
        ("0.960923", "2.2beta29", operator.lt, True),
        ("1.13++", "5.5.kw", operator.lt, True),
    ],
)
def test_loose_version_operator(v1, v2, operator, expected):
    v1 = audeer.LooseVersion(v1)
    v2 = audeer.LooseVersion(v2)
    result = operator(v1, v2)
    assert result == expected


@pytest.mark.parametrize(
    "v1, v2, wanted",
    [
        ("1.5.1", "1.5.2b2", -1),
        pytest.param(
            "161",
            "3.10a",
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        ("8.02", "8.02", 0),
        pytest.param(
            "3.4j",
            "1996.07.12",
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        pytest.param(
            "3.2.pl0",
            "3.1.1.6",
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        pytest.param(
            "2g6",
            "11g",
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        ("0.9", "2.2", -1),
        ("1.2.1", "1.2", 1),
        ("1.1", "1.2.2", -1),
        ("1.2", "1.1", 1),
        ("1.2.1", "1.2.2", -1),
        ("1.2.2", "1.2", 1),
        ("1.2", "1.2.2", -1),
        ("0.4.0", "0.4", 0),
        pytest.param(
            "1.13++",
            "5.5.kw",
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
    ],
)
def test_strict_version(v1, v2, wanted):
    # This test is replicating the original tests from
    # https://github.com/python/cpython/blob/20a1c8ee4bcb1c421b7cca1f3f5d6ad7ce30a9c9/Lib/distutils/tests/test_version.py

    v1 = audeer.StrictVersion(v1)

    result = v1._cmp(audeer.StrictVersion(v2))
    assert result == wanted

    result = v1._cmp(v2)
    assert result == wanted

    result = v1._cmp(object())
    assert result == NotImplemented

    assert v1.__repr__() == f"StrictVersion ('{str(v1)}')"


@pytest.mark.parametrize(
    "v1, v2, operator, expected",
    [
        ("1.5.1", "1.5.2b2", operator.eq, False),
        ("1.5.1", "1.5.2b2", operator.lt, True),
        ("1.5.1", "1.5.2b2", operator.le, True),
        ("1.5.1", "1.5.2b2", operator.gt, False),
        ("1.5.1", "1.5.2b2", operator.ge, False),
        ("1.5.1", "1.5.2b2", operator.ge, False),
        ("1.5.1a1", "1.5.1", operator.lt, True),
        ("1.5.1a1", "1.5.1a2", operator.lt, True),
        ("1.5.1a2", "1.5.1a1", operator.gt, True),
        ("1.5.1a1", "1.5.1a1", operator.eq, True),
        ("1.5.1", "1.5.1b2", operator.gt, True),
        ("1.5.1", "1.5.1", operator.le, True),
        ("1.5.1", "1.5.1", operator.ge, True),
        ("8.02", "8.02", operator.eq, True),
        ("0.9", "2.2", operator.lt, True),
        ("1.2.1", "1.2", operator.gt, True),
        ("1.1", "1.2.2", operator.lt, True),
        ("1.2", "1.1", operator.gt, True),
        ("1.2.1", "1.2.2", operator.lt, True),
        ("1.2.2", "1.2", operator.gt, True),
        ("1.2", "1.2.2", operator.lt, True),
        ("0.4.0", "0.4", operator.eq, True),
    ],
)
def test_strict_version_operator(v1, v2, operator, expected):
    v1 = audeer.StrictVersion(v1)
    v2 = audeer.StrictVersion(v2)
    result = operator(v1, v2)
    assert result == expected



class config:
    """Get/set defaults for :mod:`audeer`.

    For example, when you want to change the default number of columns
    for the progress bar::

        import audeer

        audeer.config.TQDM_COLUMNS = 50

    """

    TQDM_DESCLEN = 60
    """Length of progress bar description."""

    TQDM_FORMAT = (
        "{percentage:3.0f}%|{bar} [{elapsed}<{remaining}] "
        "{desc:" + str(TQDM_DESCLEN) + "}"
    )
    """Format of progress bars."""

    TQDM_COLUMNS = 100
    """Number of columns of progress bars."""

    TQDM_LEAVE = False
    """Leave progress bar on screen after finishing."""
